# coding=utf-8
# Copyright (c) 2014, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module providing tests for the core module """

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import collections
import errno
import os
import textwrap

import pytest
import six

from framework import core
from framework import exceptions

from . import skip

# Making good test names often flies in the face of PEP8 recomendations, ignore
# those
# pylint: disable=invalid-name
#
# For testing purposes (and to get good names) it's useful to use classes for
# grouping tests. Often these tests share no state, which would be bad if they
# were not tests, but that's kind of the way python testing works.
# pylint: disable=no-self-use


class TestParseListfile(object):
    """Tests for core.parse_listfile."""

    def test_parse_listfile_return(self, tmpdir):
        """core.parse_listfile(): returns a list-like object.

        Given a file with a newline separated list of results, parse_listfile
        should return a list of files with no whitespace.
        """
        f = tmpdir.join('test.list')
        f.write("/tmp/foo\n/tmp/bar\n")
        results = core.parse_listfile(six.text_type(f))
        assert isinstance(results, collections.Container)

    def test_parse_listfile_whitespace(self, tmpdir):
        """parse_listfile should remove various kinds of trailing whitespace.

        It is important it doesn't touch whitespace in lines, however.
        """
        f = tmpdir.join('test.list')
        f.write(textwrap.dedent("""\
            space between
            tab\t
            space
            newline
        """))
        results = core.parse_listfile(six.text_type(f))

        assert results[0] == 'space between'
        assert results[1] == 'tab'
        assert results[2] == 'space'
        assert results[3] == 'newline'

    def test_parse_listfile_tilde_posix(self, tmpdir):
        """core.parse_listfile(): tildes (~) are properly expanded.

        According to the python docs for python 2.7
        (http://docs.python.org/2/library/os.path.html#module-os.path), both
        os.path.expanduser and os.path.expandvars work on both *nix systems
        (Linux, *BSD, OSX) and Windows.
        """
        if os.name == 'posix':
            expected = os.path.expandvars('$HOME')
        else:
            expected = os.path.expandvars('%USERPROFILE%')
        expected = os.path.normpath(os.path.join(expected, 'foo'))

        f = tmpdir.join('test.list')
        f.write("~/foo\n")
        results = core.parse_listfile(six.text_type(f))
        assert os.path.normpath(results[0]) == expected


class TestGetConfig(object):
    """Tests for core.get_config."""
    _CONF_FILE = textwrap.dedent("""\
        [nose-test]
        ; a section for testing behavior
        dir = foo""")

    @skip.linux
    def test_config_in_xdg_config_home(self, tmpdir, mocker):
        """core.get_config() finds $XDG_CONFIG_HOME/piglit.conf"""
        env = mocker.patch('framework.core.os.environ', new={})
        env['XDG_CONFIG_HOME'] = six.text_type(tmpdir)
        conf = tmpdir.join('piglit.conf')
        conf.write(self._CONF_FILE)
        core.get_config()

        assert core.PIGLIT_CONFIG.has_section('nose-test')

    @skip.linux
    def test_config_in_home_dir(self, tmpdir, mocker):
        """core.get_config() finds $HOME/.config/piglit.conf"""
        env = mocker.patch('framework.core.os.environ', new={})
        env['HOME'] = six.text_type(tmpdir)
        conf = tmpdir.join('piglit.conf')
        conf.write(self._CONF_FILE)
        core.get_config()

        assert core.PIGLIT_CONFIG.has_section('nose-test')

    def test_config_in_current(self, tmpdir, mocker):
        """core.get_config() finds ./piglit.conf"""
        mocker.patch('framework.core.os.environ', new={})
        conf = tmpdir.join('piglit.conf')
        conf.write(self._CONF_FILE)

        ret = tmpdir.chdir()
        try:
            core.get_config()
        finally:
            os.chdir(six.text_type(ret))

        assert core.PIGLIT_CONFIG.has_section('nose-test')

    def test_config_in_piglit_root(self, mocker, tmpdir):
        """core.get_config() finds "piglit root"/piglit.conf"""
        # Mock the __file__ attribute of the core module, since that's how
        # piglit decides where the root of the piglit directory is.
        mocker.patch('framework.core.__file__',
                     six.text_type(tmpdir.join('framework', 'core.py')))
        mocker.patch('framework.core.os.environ', new={})
        conf = tmpdir.join('piglit.conf')
        conf.write(self._CONF_FILE)
        core.get_config()

        assert core.PIGLIT_CONFIG.has_section('nose-test')


class TestPiglitConfig(object):
    """Tests for PiglitConfig methods."""
    @classmethod
    def setup_class(cls):
        cls.conf = core.PiglitConfig()
        cls.conf.add_section('set')
        cls.conf.set('set', 'options', 'bool')

    def test_safe_get_valid(self):
        """core.PiglitConfig: safe_get returns a value if its in the Config."""
        assert self.conf.safe_get('set', 'options') == 'bool'

    def test_PiglitConfig_required_get_valid(self):
        """core.PiglitConfig: required_get returns a value if its in the
        Config."""
        assert self.conf.required_get('set', 'options') == 'bool'

    def test_safe_get_missing_option(self):
        """core.PiglitConfig: safe_get returns None if the option is missing.
        """
        assert self.conf.safe_get('set', 'invalid') is None

    def test_safe_get_missing_section(self):
        """core.PiglitConfig: safe_get returns None if the section is missing.
        """
        assert self.conf.safe_get('invalid', 'invalid') is None

    def test_required_get_missing_option(self):
        """core.PiglitConfig: required_get raises PiglitFatalError if the
        option is missing."""
        with pytest.raises(exceptions.PiglitFatalError):
            self.conf.required_get('set', 'invalid')

    def test_required_get_missing_section(self):
        """core.PiglitConfig: required_get raises PiglitFatalError if the
        section is missing."""
        with pytest.raises(exceptions.PiglitFatalError):
            self.conf.required_get('invalid', 'invalid')

    def test_safe_get_fallback(self):
        """core.PiglitConfig: safe_get returns the value of fallback when the
        section or option is missing."""
        assert self.conf.safe_get('invalid', 'invalid', fallback='foo') == 'foo'


class TestCheckDir(object):
    """Tests for core.check_dir."""

    def test_exists_fail(self, mocker, tmpdir):
        """core.check_dir: if the directory exists and failifexsits is True
        fail."""
        tmpdir.chdir()
        mocker.patch('framework.core.os.stat', mocker.Mock(side_effect=OSError))
        with pytest.raises(exceptions.PiglitException):
            core.check_dir('foo', True)

    def test_not_exists_and_not_fail(self, mocker, tmpdir):
        """core.check_dir: if the directory doesn't exists (ENOENT) and
        failifexists is False continue."""
        tmpdir.chdir()
        mocker.patch('framework.core.os.stat',
                     mocker.Mock(side_effect=OSError('foo', errno.ENOENT)))
        makedirs = mocker.patch('framework.core.os.makedirs')

        core.check_dir('foo', False)

        assert makedirs.called == 1

    def test_exists_and_not_fail(self, mocker, tmpdir):
        """core.check_dir: If makedirs fails with EEXIST pass"""
        tmpdir.chdir()
        mocker.patch('framework.core.os.stat', mocker.Mock())
        makedirs = mocker.patch(
            'framework.core.os.makedirs',
            mocker.Mock(side_effect=OSError(errno.EEXIST, 'foo')))

        core.check_dir('foo', False)

        assert makedirs.called == 0

    def test_makedirs_fail(self, mocker, tmpdir):
        """core.check_dir: If makedirs fails with any other raise that error."""
        tmpdir.chdir()
        mocker.patch('framework.core.os.makedirs',
                     mocker.Mock(side_effect=OSError))

        with pytest.raises(OSError):
            core.check_dir('foo', False)

    def test_handler(self, mocker, tmpdir):
        """core.check_dir: Handler is called if not failifexists."""
        class Sentinel(Exception):
            pass

        tmpdir.chdir()

        mocker.patch('framework.core.os.stat',
                     mocker.Mock(side_effect=OSError('foo', errno.ENOTDIR)))
        with pytest.raises(Sentinel):
            core.check_dir('foo', handler=mocker.Mock(side_effect=Sentinel))

    @skip.PY2
    def test_stat_FileNotFoundError(self, mocker, tmpdir):
        """core.check_dir: FileNotFoundError is raised and failifexsits is
        False continue."""
        tmpdir.chdir()
        mocker.patch('framework.core.os.stat',
                     mocker.Mock(side_effect=FileNotFoundError))
        makedirs = mocker.patch('framework.core.os.makedirs')

        core.check_dir('foo', False)

        assert makedirs.called == 1



# coding=utf-8
#
# Copyright © 2016 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Generate fp64 vertex shader input tests."""

from __future__ import print_function, division, absolute_import
import abc
import argparse
import itertools
import os
import types

from six.moves import range

from templates import template_dir
from modules import utils
from modules import types as glsltypes

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

# Hard limit so we don't generate useless tests that cannot be run in any existing HW.
MAX_VERTEX_ATTRIBS = 32

# pylint: disable=bad-whitespace,line-too-long
DOUBLE_INFS            = ['0xfff0000000000000',  # -inf
                          '0x7ff0000000000000']  # +inf

DOUBLE_NEG_ZERO        = ['0x8000000000000000']  # Negative underflow (-0.0)

DOUBLE_POS_ZERO        = ['0x0000000000000000']  # Positive underflow (+0.0)

# Double values causing an underflow to zero in any other type
DOUBLE_DENORMAL_VALUES = ['0x800fffffffffffff',  # Negative maximum denormalized -- Denormalized may be flushed to 0
                          '0x8000000000000001',  # Negative minimum denormalized -- Denormalized may be flushed to 0
                          '0x0000000000000001',  # Positive minimum denormalized -- Denormalized may be flushed to 0
                          '0x000fffffffffffff']  # Positive maximum denormalized -- Denormalized may be flushed to 0

DOUBLE_NORMAL_VALUES   = ['0xffefffffffffffff',  # Negative maximum normalized
                          '0xcb1e35ed24eb6496',  # -7.23401345e+53
                          '0xc8b1381a93a87849',  # -1.5e+42
                          '0xc7efffffefffffff',  # Negative maximum float normalized
                          '0xc170000000000000',  # -16777216.0
                          '0xc014000000000000',  # -5.0
                          '0xbfff25ce60000000',  # -1.9467300176620483
                          '0x8010000000000000',  # Negative minimum normalized
                          '0x0010000000000000',  # Positive minimum normalized
                          '0x3fff25ce60000000',  # +1.9467300176620483
                          '0x4014000000000000',  # +5.0
                          '0x4170000000000000',  # +16777216.0
                          '0x47efffffefffffff',  # Positive maximum float normalized
                          '0x48b1381a93a87849',  # +1.5e+42
                          '0x4b1e35ed24eb6496',  # +7.23401345e+53
                          '0x7fefffffffffffff']  # Positive maximum normalized

FLOAT_POS_ZERO         = ['0x00000000']  #  0.0

FLOAT_NORMAL_VALUES    = ['0xc21620c5',  # -3.7532
                          '0x75bc289b',  #  4.7703e32
                          '0x54c1c081',  #  6.6572e12
                          '0x878218f8',  # -1.9575e-34
                          '0x7e0857ed',  #  4.5307886e37
                          '0x2bb561bf',  #  1.2887954e-12
                          '0xff7fffff',  # Negative maximum normalized
                          '0xcb800000',  # -16777216.0
                          '0xc0a00000',  # -5.0
                          '0xbff92e73',  # -1.9467300
                          '0x80800000',  # Negative minimum normalized
                          '0x00800000',  # Positive minimum normalized
                          '0x3ff92e73',  #  1.9467300
                          '0x40a00000',  #  5.0
                          '0x4b800000',  #  16777216.0
                          '0x7f7fffff']  # Positive maximum normalized

UBYTE_VALUES           = ['0',    # Minimum
                          '127',  # Signed byte low frontier
                          '128',  # Signed byte up frontier
                          '255',  # Maximum
                          '1',
                          '5',
                          '14',
                          '23',
                          '58',
                          '91',
                          '113',
                          '135',
                          '179',
                          '185',
                          '205',
                          '207',
                          '212']

BYTE_VALUES            = ['-128',  # Minimum
                          '-5',
                          '-1',
                          '0',
                          '1',
                          '5',
                          '127',   # Maximum
                          '-125',
                          '-120',
                          '-117',
                          '-69',
                          '-24',
                          '-20',
                          '21',
                          '89',
                          '106',
                          '119']

USHORT_VALUES          = ['0',      # Minimum
                          '32767',  # Signed short low frontier
                          '32768',  # Signed short up frontier
                          '65535',  # Maximum
                          '1',
                          '5',
                          '12610',
                          '17110',
                          '19962',
                          '23589',
                          '37265',
                          '41792',
                          '45699',
                          '47934',
                          '55916',
                          '56412',
                          '65142']

SHORT_VALUES           = ['-32768',  # Minimum
                          '-5',
                          '-1',
                          '0',
                          '1',
                          '5',
                          '32767',   # Maximum
                          '-16255',
                          '-12480',
                          '-12360',
                          '1706',
                          '5386',
                          '7315',
                          '18137',
                          '25871',
                          '26255',
                          '26472']

UINT_VALUES            = ['0',           # Minimum
                          '2147483647',  # Signed int low frontier
                          '2147483648',  # Signed int up frontier
                          '4294967295',  # Maximum
                          '1',
                          '5',
                          '1073294963',
                          '1084227584',
                          '1266679808',
                          '1421983873',
                          '2114476013',
                          '2273450232',
                          '3220778611',
                          '3231711232',
                          '3256230085',
                          '3414163456',
                          '4294967294']

INT_VALUES             = ['-2147483648',  # Minimum
                          '-5',
                          '-1',
                          '0',
                          '1',
                          '5',
                          '2147483647',   # Maximum
                          '-1038737211',
                          '-1063256064',
                          '-1074188685',
                          '-880803840',
                          '1073294963',
                          '1084227584',
                          '1266679808',
                          '1421983873',
                          '1975265435',
                          '2114476013']

GL_TYPES_VALUES        = {'double': DOUBLE_NORMAL_VALUES + DOUBLE_POS_ZERO,
                          'float': FLOAT_NORMAL_VALUES + FLOAT_POS_ZERO,
                          'ubyte': UBYTE_VALUES,
                          'byte': BYTE_VALUES,
                          'ushort': USHORT_VALUES,
                          'short': SHORT_VALUES,
                          'uint': UINT_VALUES,
                          'int': INT_VALUES}

GLSL_DSCALAR_TYPES     = [glsltypes.DOUBLE]

GLSL_DVEC_TYPES        = [glsltypes.DVEC2, glsltypes.DVEC3, glsltypes.DVEC4]

GLSL_DMAT_TYPES        = [glsltypes.DMAT2, glsltypes.DMAT2X3, glsltypes.DMAT2X4,
                          glsltypes.DMAT3X2, glsltypes.DMAT3, glsltypes.DMAT3X4,
                          glsltypes.DMAT4X2, glsltypes.DMAT4X3, glsltypes.DMAT4]

GLSL_FSCALAR_TYPES     = [glsltypes.FLOAT]

GLSL_FVEC_TYPES        = [glsltypes.VEC2, glsltypes.VEC3, glsltypes.VEC4]

GLSL_FMAT_TYPES        = [glsltypes.MAT2, glsltypes.MAT2X3, glsltypes.MAT2X4,
                          glsltypes.MAT3X2, glsltypes.MAT3, glsltypes.MAT3X4,
                          glsltypes.MAT4X2, glsltypes.MAT4X3, glsltypes.MAT4]

GLSL_ISCALAR_TYPES     = [glsltypes.INT]

GLSL_IVEC_TYPES        = [glsltypes.IVEC2, glsltypes.IVEC3, glsltypes.IVEC4]

GLSL_USCALAR_TYPES     = [glsltypes.UINT]

GLSL_UVEC_TYPES        = [glsltypes.UVEC2, glsltypes.UVEC3, glsltypes.UVEC4]


# pylint: enable=bad-whitespace,line-too-long


class TestTuple(object):
    """A float64 derived and other type derived tuple to generate the
       needed conversion tests.
    """
    @staticmethod
    def get_dir_name(ver):
        """Returns the directory name to save tests given a GLSL version."""

        assert isinstance(ver, str)
        if ver.startswith('GL_'):
            feature_dir = ver[3:].lower()
        else:
            feature_dir = 'glsl-{}.{}'.format(ver[0], ver[1:])

        return os.path.join('spec', feature_dir, 'execution',
                            'vs_in')

    def __init__(self, ver, names_only):
        assert isinstance(ver, str)
        assert isinstance(names_only, bool)

        self._ver = ver
        self._names_only = names_only

    @abc.abstractmethod
    def generate(self):
        """Generate the GLSL parser tests."""


class RegularTestTuple(TestTuple):
    """Derived class for conversion tests using regular values within the
       edges of the used types.
    """

    @staticmethod
    def create_in_types_array(*types_arrays):
        """Creates vertex input combinations."""

        for product_item in itertools.product(*types_arrays):
            yield product_item

    @staticmethod
    def create_tests(glsl_vers, in_types_array, gl_types,
                     position_orders, arrays_array, names_only):
        """Creates combinations for flat qualifier tests."""

        assert isinstance(glsl_vers, list)
        assert isinstance(in_types_array, types.GeneratorType)
        assert isinstance(gl_types, list)
        assert isinstance(position_orders, list)
        assert isinstance(arrays_array, list)
        assert isinstance(names_only, bool)

        if not names_only:
            for ver in glsl_vers:
                utils.safe_makedirs(TestTuple.get_dir_name(ver))

        for in_types, position_order, arrays, ver in itertools.product(
                in_types_array,
                position_orders,
                arrays_array,
                glsl_vers):
            num_vs_in = 1  # We use an additional vec3 piglit_vertex input
            for idx, in_type in enumerate(in_types):
                num_vs_in += (in_type.columns or 1) * arrays[idx] * \
                    (2 if in_type.type.name == 'double' and in_type.rows in [3, 4] else 1)
                # dvec* and dmat* didn't appear in GLSL until 4.20
                if (in_type.type.name == 'double' and not in_type.scalar) and ver == '410':
                    ver = '420'
            # Skip the test if it needs too many inputs
            if num_vs_in > MAX_VERTEX_ATTRIBS:
                continue

            yield ver, in_types, gl_types, position_order, arrays, num_vs_in, names_only

    @staticmethod
    def all_tests(names_only):
        """Creates all the combinations for flat qualifier tests."""

        assert isinstance(names_only, bool)

        # We need additional directories for GLSL 420
        if not names_only:
            utils.safe_makedirs(TestTuple.get_dir_name('420'))
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_USCALAR_TYPES, GLSL_UVEC_TYPES),
                    itertools.chain(GLSL_ISCALAR_TYPES, GLSL_IVEC_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['ubyte', 'short', 'double'],
                [1, 2, 3, 4],
                [[1, 1, 1]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_ISCALAR_TYPES, GLSL_IVEC_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['byte', 'double'],
                [1, 2, 3],
                [[1, 1]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_USCALAR_TYPES, GLSL_UVEC_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['ushort', 'double'],
                [1, 2, 3],
                [[1, 1]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES),
                    itertools.chain(GLSL_FSCALAR_TYPES, GLSL_FVEC_TYPES, GLSL_FMAT_TYPES)),
                ['double', 'float'],
                [1, 2, 3],
                [[1, 1], [1, 3], [5, 1], [5, 3]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES),
                    itertools.chain(GLSL_ISCALAR_TYPES, GLSL_IVEC_TYPES)),
                ['double', 'int'],
                [1, 2, 3],
                [[1, 1], [1, 3], [5, 1], [5, 3]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES),
                    itertools.chain(GLSL_USCALAR_TYPES, GLSL_UVEC_TYPES)),
                ['double', 'uint'],
                [1, 2, 3],
                [[1, 1], [1, 3], [5, 1], [5, 3]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_FSCALAR_TYPES, GLSL_FVEC_TYPES, GLSL_FMAT_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['float', 'double'],
                [1, 2, 3],
                [[1, 1], [1, 2], [3, 1], [3, 2]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_ISCALAR_TYPES, GLSL_IVEC_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['int', 'double'],
                [1, 2, 3],
                [[1, 1], [1, 2], [3, 1], [3, 2]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_USCALAR_TYPES, GLSL_UVEC_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['uint', 'double'],
                [1, 2, 3],
                [[1, 1], [1, 2], [3, 1], [3, 2]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES),
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['double', 'double'],
                [1, 2, 3],
                [[1, 1], [1, 2], [3, 1], [3, 2]],
                names_only):
            yield RegularTestTuple(*test_args)
        for test_args in RegularTestTuple.create_tests(
                ['GL_ARB_vertex_attrib_64bit', '410'],
                RegularTestTuple.create_in_types_array(
                    itertools.chain(GLSL_DSCALAR_TYPES, GLSL_DVEC_TYPES, GLSL_DMAT_TYPES)),
                ['double'],
                [1, 2],
                [[1], [5]],
                names_only):
            yield RegularTestTuple(*test_args)

    def __init__(self, ver, in_types, gl_types, position_order, arrays, num_vs_in, names_only):
        assert ver in ('GL_ARB_vertex_attrib_64bit', '410', '420')
        assert isinstance(in_types, tuple)
        assert isinstance(gl_types, list)
        assert len(gl_types) == len(in_types)
        assert isinstance(position_order, int)
        assert (position_order > 0) and (position_order - 1 <= len(in_types))
        assert isinstance(arrays, list)
        assert isinstance(num_vs_in, int) and (num_vs_in <= MAX_VERTEX_ATTRIBS)
        super(RegularTestTuple, self).__init__(ver, names_only)

        self._in_types = in_types
        self._gl_types = gl_types
        self._position_order = position_order
        self._arrays = arrays
        self._num_vs_in = num_vs_in

    def generate(self):
        """Generate GLSL parser tests."""
        filename = os.path.join(TestTuple.get_dir_name(self._ver), 'vs-input')
        for idx, in_type in enumerate(self._in_types):
            if idx == self._position_order - 1:
                filename += '-position'
            filename += '-{}_{}{}'.format(
                self._gl_types[idx], in_type.name, '_array{}'.format(
                    self._arrays[idx]) if self._arrays[idx] - 1 else '')
        if self._position_order > len(self._in_types):
            filename += '-position'
        filename += '.shader_test'

        if not self._names_only:
            with open(filename, 'w') as test_file:
                test_file.write(TEMPLATES.get_template(
                    'regular.shader_test.mako').render_unicode(
                        ver=self._ver,
                        in_types=self._in_types,
                        gl_types=self._gl_types,
                        position_order=self._position_order,
                        arrays=self._arrays,
                        num_vs_in=self._num_vs_in,
                        gl_types_values=GL_TYPES_VALUES))

        print(filename)


class ColumnsTestTuple(TestTuple):
    """Derived class for conversion tests using regular values within the
       edges of the used types.
    """

    @staticmethod
    def all_tests(names_only):
        """Creates all the combinations for flat qualifier tests."""

        assert isinstance(names_only, bool)
        glsl_vers = ['GL_ARB_vertex_attrib_64bit', '420']

        if not names_only:
            for ver in glsl_vers:
                utils.safe_makedirs(TestTuple.get_dir_name(ver))

        for mat in GLSL_DMAT_TYPES:
            for columns in itertools.product(range(2), repeat=mat.columns):
                if (0 not in columns) or (1 not in columns):
                    continue
                for ver in glsl_vers:
                    yield ColumnsTestTuple(ver, mat, columns, names_only)

    def __init__(self, ver, mat, columns, names_only):
        assert ver in ('GL_ARB_vertex_attrib_64bit', '420')
        super(ColumnsTestTuple, self).__init__(ver, names_only)

        self._mat = mat
        self._columns = columns

    def generate(self):
        """Generate GLSL parser tests."""

        filename = os.path.join(TestTuple.get_dir_name(self._ver),
                                'vs-input-columns-{}'.format(self._mat.name))
        for idx, column in enumerate(self._columns):
            if column == 1:
                filename += '-{}'.format(idx)
        filename += '.shader_test'

        if not self._names_only:
            with open(filename, 'w') as test_file:
                test_file.write(TEMPLATES.get_template(
                    'columns.shader_test.mako').render_unicode(
                        ver=self._ver,
                        mat=self._mat,
                        columns=self._columns,
                        dvalues=GL_TYPES_VALUES['double']))

        print(filename)


def main():
    """Main function."""

    parser = argparse.ArgumentParser(
        description="Generate non-flat interpolation qualifier tests with fp64 types")
    parser.add_argument(
        '--names-only',
        dest='names_only',
        action='store_true',
        default=False,
        help="Don't output files, just generate a list of filenames to stdout")
    args = parser.parse_args()

    for test in itertools.chain(RegularTestTuple.all_tests(args.names_only),
                                ColumnsTestTuple.all_tests(args.names_only)):
        test.generate()


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2015, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Shared constants for test modules."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os
import sys
import subprocess

TESTS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

# If the PIGLIT_BUILD_DIR env var is set, we want to look for the
# generated tests in $PIGLIT_BUILD_DIR/generated_tests/.  Otherwise,
# look in TESTS_DIR/../generated_tests.
build_dir = os.environ.get('PIGLIT_BUILD_TREE')
if not build_dir:
    build_dir = os.environ.get('PIGLIT_BUILD_DIR', os.path.join(TESTS_DIR, '..'))

GENERATED_TESTS_DIR = os.path.abspath(os.path.join(build_dir, 'generated_tests'))

# If on cygwin convert to a dos style path
if sys.platform == 'cygwin':
    def dosify(p):
        return subprocess.check_output(
            ['cygpath', '-d', p]).rstrip().decode('utf-8')

    TESTS_DIR = dosify(TESTS_DIR)
    GENERATED_TESTS_DIR = dosify(GENERATED_TESTS_DIR)



# coding=utf-8
# Copyright (c) 2014-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for JSON backend version updates."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os
try:
    import simplejson as json
except ImportError:
    import json
try:
    import mock
except ImportError:
    from unittest import mock

import jsonschema
import pytest

from framework import backends

# pylint: disable=protected-access,no-self-use


@pytest.yield_fixture(autouse=True, scope='module')
def setup_module():
    with mock.patch.dict(backends.json.compression.os.environ,
                         {'PIGLIT_COMPRESSION': 'none'}):
        yield


class TestV7toV8(object):
    """Tests for Version 7 to version 8."""

    data = {
        "results_version": 7,
        "name": "test",
        "options": {
            "profile": ['quick'],
            "dmesg": False,
            "verbose": False,
            "platform": "gbm",
            "sync": False,
            "valgrind": False,
            "filter": [],
            "concurrent": "all",
            "test_count": 0,
            "exclude_tests": [],
            "exclude_filter": [],
            "env": {},
        },
        "lspci": "stuff",
        "uname": "more stuff",
        "glxinfo": "and stuff",
        "wglinfo": "stuff",
        "clinfo": "stuff",
        "tests": {
            'a@test': {
                'time': 1.2,
                'dmesg': '',
                'result': 'fail',
                '__type__': 'TestResult',
                'command': '/a/command',
                'traceback': None,
                'out': '',
                'environment': 'A=variable',
                'returncode': 0,
                'err': '',
                'pid': 5,
                'subtests': {
                    '__type__': 'Subtests',
                },
                'exception': None,
            }
        },
        "time_elapsed": 1.2,
        '__type__': 'TestrunResult',
    }

    @pytest.fixture
    def result(self, tmpdir):
        p = tmpdir.join('result.json')
        p.write(json.dumps(self.data, default=backends.json.piglit_encoder))
        with p.open('r') as f:
            return backends.json._update_seven_to_eight(backends.json._load(f))

    def test_time(self, result):
        """backends.json.update_results (7 -> 8): test time is stored as start
        and end.
        """
        assert result['tests']['a@test']['time']['start'] == 0.0
        assert result['tests']['a@test']['time']['end'] == 1.2

    def test_time_elapsed(self, result):
        """backends.json.update_results (7 -> 8): total time is stored as start
        and end.
        """
        assert result['time_elapsed']['start'] == 0.0
        assert result['time_elapsed']['end'] == 1.2

    def test_valid(self, result):
        with open(os.path.join(os.path.dirname(__file__), 'schema',
                               'piglit-8.json'),
                  'r') as f:
            schema = json.load(f)
        jsonschema.validate(
            json.loads(json.dumps(result, default=backends.json.piglit_encoder)),
            schema)


class TestV8toV9(object):
    """Tests for Version 8 to version 9."""

    data = {
        "results_version": 8,
        "name": "test",
        "options": {
            "profile": ['quick'],
            "dmesg": False,
            "verbose": False,
            "platform": "gbm",
            "sync": False,
            "valgrind": False,
            "filter": [],
            "concurrent": "all",
            "test_count": 0,
            "exclude_tests": [],
            "exclude_filter": [],
            "env": {},
        },
        "lspci": "stuff",
        "uname": "more stuff",
        "glxinfo": "and stuff",
        "wglinfo": "stuff",
        "clinfo": "stuff",
        "tests": {
            'a@test': {
                "time": {
                    'start': 1.2,
                    'end': 1.8,
                    '__type__': 'TimeAttribute'
                },
                'dmesg': '',
                'result': 'fail',
                '__type__': 'TestResult',
                'command': '/a/command',
                'traceback': None,
                'out': '',
                'environment': 'A=variable',
                'returncode': 0,
                'err': '',
                'pid': 5,
                'subtests': {
                    '__type__': 'Subtests',
                },
                'exception': None,
            },
            'b@test': {
                "time": {
                    'start': 1.2,
                    'end': 1.8,
                    '__type__': 'TimeAttribute'
                },
                'dmesg': '',
                'result': 'fail',
                '__type__': 'TestResult',
                'command': '/a/command',
                'traceback': None,
                'out': '',
                'environment': 'A=variable',
                'returncode': 0,
                'err': '',
                'subtests': {
                    '__type__': 'Subtests',
                },
                'exception': None,
            }
        },
        "time_elapsed": {
            'start': 1.2,
            'end': 1.8,
            '__type__': 'TimeAttribute'
        },
        '__type__': 'TestrunResult',
    }

    @pytest.fixture
    def result(self, tmpdir):
        p = tmpdir.join('result.json')
        p.write(json.dumps(self.data, default=backends.json.piglit_encoder))
        with p.open('r') as f:
            return backends.json._update_eight_to_nine(backends.json._load(f))

    def test_pid(self, result):
        assert result['tests']['a@test']['pid'] == [5]

    def test_no_pid(self, result):
        assert result['tests']['b@test']['pid'] == []

    def test_valid(self, result):
        with open(os.path.join(os.path.dirname(__file__), 'schema',
                               'piglit-9.json'),
                  'r') as f:
            schema = json.load(f)
        jsonschema.validate(
            json.loads(json.dumps(result, default=backends.json.piglit_encoder)),
            schema)


class TestV9toV10(object):
    """Tests for Version 8 to version 9."""

    data = {
        "results_version": 9,
        "name": "test",
        "options": {
            "profile": ['quick'],
            "dmesg": False,
            "verbose": False,
            "platform": "gbm",
            "sync": False,
            "valgrind": False,
            "filter": [],
            "concurrent": "all",
            "test_count": 0,
            "exclude_tests": [],
            "exclude_filter": [],
            "env": {},
        },
        "lspci": "stuff",
        "uname": "more stuff",
        "glxinfo": "and stuff",
        "wglinfo": "stuff",
        "clinfo": "stuff",
        "tests": {
            'a@test': {
                "time": {
                    'start': 1.2,
                    'end': 1.8,
                    '__type__': 'TimeAttribute'
                },
                'dmesg': '',
                'result': 'fail',
                '__type__': 'TestResult',
                'command': '/a/command',
                'traceback': None,
                'out': '',
                'environment': 'A=variable',
                'returncode': 0,
                'err': '',
                'pid': [5],
                'subtests': {
                    '__type__': 'Subtests',
                },
                'exception': None,
            },
        },
        "time_elapsed": {
            'start': 1.2,
            'end': 1.8,
            '__type__': 'TimeAttribute'
        },
        '__type__': 'TestrunResult',
    }

    @pytest.fixture
    def result(self, tmpdir):
        p = tmpdir.join('result.json')
        p.write(json.dumps(self.data, default=backends.json.piglit_encoder))
        with p.open('r') as f:
            return backends.json._update_nine_to_ten(backends.json._load(f))

    @pytest.mark.parametrize("key", ['glxinfo', 'wglinfo', 'clinfo', 'uname', 'lspci'])
    def test(self, key, result):
        assert key not in result, 'Root key/value not removed'
        assert key in result['info']['system'], 'Key not added to info/system'
        assert result['info']['system'][key] == self.data[key], \
            'Value not set properly.'

    def test_valid(self, result):
        with open(os.path.join(os.path.dirname(__file__), 'schema',
                               'piglit-10.json'),
                  'r') as f:
            schema = json.load(f)
        jsonschema.validate(
            json.loads(json.dumps(result, default=backends.json.piglit_encoder)),
            schema)



# coding=utf-8
#
# Copyright © 2011 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# Generate a set of shader_runner tests for every overloaded version
# of every built-in function, based on the test vectors computed by
# builtin_function.py.
#
# In each set of generated tests, one test exercises the built-in
# function in each type of shader (vertex, geometry, and fragment).
# In all cases, the inputs to the built-in function come from
# uniforms, so that the effectiveness of the test won't be
# circumvented by constant folding in the GLSL compiler.
#
# The tests operate by invoking the built-in function in the
# appropriate shader, applying a scale and offset so that the expected
# values are in the range [0.25, 0.75], and then outputting the result
# as a solid rgba color, which is then checked using shader_runner's
# "probe rgba" command.
#
# For built-in functions whose result type is a matrix, the test
# checks one column at a time.
#
# This program outputs, to stdout, the name of each file it generates.
# With the optional argument --names-only, it only outputs the names
# of the files; it doesn't generate them.

from __future__ import print_function, division, absolute_import
from builtin_function import *
import abc
import numpy
import optparse
import os
import os.path

from six.moves import range

from modules import utils


def compute_offset_and_scale(test_vectors):
    """Compute scale and offset values such that for each result in
    test_vectors, (result - offset) * scale is in the range [0.25,
    0.75], and scale is less than or equal to 1.0.  These values are
    used to transform the test vectors so that their outputs can be
    stored in gl_FragColor without overflow.
    """
    low = min(numpy.min(tv.result) for tv in test_vectors)
    hi = max(numpy.max(tv.result) for tv in test_vectors)
    span = hi - low
    center = (hi + low)/2.0
    span *= 2.0
    if span < 1.0:
        span = 1.0
    offset = center - span/2.0
    scale = 1.0/span
    return offset, scale


def shader_runner_format(values):
    """Format the given values for use in a shader_runner "uniform" or
    "probe rgba" command.  Bools are converted to 0's and 1's, and
    values are separated by spaces.
    """
    transformed_values = []
    for value in values:
        if isinstance(value, (bool, np.bool_)):
            transformed_values.append(int(value))
        else:
            transformed_values.append(value)
    return ' '.join(repr(x) for x in transformed_values)


def shader_runner_type(glsl_type):
    """Return the appropriate type name necessary for binding a
    uniform of the given type using shader_runner's "uniform" command.
    Boolean values and vectors are converted to ints, and square
    matrices are written in "matNxN" form.
    """
    if glsl_type.base_type == glsl_bool:
        if glsl_type.is_scalar:
            return 'int'
        else:
            return 'ivec{0}'.format(glsl_type.num_rows)
    elif glsl_type.is_matrix:
        return 'mat{0}x{1}'.format(glsl_type.num_cols, glsl_type.num_rows)
    else:
        return str(glsl_type)


class Comparator(object):
    """Base class which abstracts how we compare expected and actual
    values.
    """
    __metaclass__ = abc.ABCMeta

    def make_additional_declarations(self):
        """Return additional declarations, if any, that are needed in
        the shader program.
        """
        return ''

    @abc.abstractmethod
    def make_result_handler(self, invocation, output_var):
        """Return the shader code that is needed to produce the result
        and store it in output_var.

        invocation is the GLSL code to compute the output of the
        built-in function.
        """

    @abc.abstractmethod
    def draw_test(self, test_vector, draw_command):
        """Return the shader_runner test code that is needed to run a
        single test vector.
        """

    @abc.abstractmethod
    def result_vector(self, test_vector):
        """Return the expected result color as a list of floats."""

    def testname_suffix(self):
        """Return a string to be used as a suffix on the test name to
        distinguish it from tests using other comparators."""
        return ''


class BoolComparator(Comparator):
    """Comparator that tests functions returning bools and bvecs by
    converting them to floats.

    This comparator causes code to be generated in the following form:

        rettype result = func(args);
        output_var = vec4(result, 0.0, ...);
    """
    def __init__(self, signature):
        assert not signature.rettype.is_matrix
        self.__signature = signature
        self.__padding = 4 - signature.rettype.num_rows

    def make_result_handler(self, invocation, output_var):
        statements = '  {0} result = {1};\n'.format(
            self.__signature.rettype, invocation)
        statements += '  {0} = vec4(result{1});\n'.format(
            output_var, ', 0.0' * self.__padding)
        return statements

    def convert_to_float(self, value):
        """Convert the given vector or scalar value to a list of
        floats representing the expected color produced by the test.
        """
        value = value*1.0  # convert bools to floats
        value = column_major_values(value)
        value += [0.0] * self.__padding
        return value

    def draw_test(self, test_vector, draw_command):
        return draw_command

    def result_vector(self, test_vector):
        return self.convert_to_float(test_vector.result)


class BoolIfComparator(Comparator):
    """Comparator that tests functions returning bools by evaluating
    them inside an if statement.

    This comparator causes code to be generated in the following form:

        if (func(args))
          output_var = vec4(1.0, 1.0, 0.0, 1.0);
        else
          output_var = vecp(0.0, 0.0, 1.0, 1.0);
    """
    def __init__(self, signature):
        assert signature.rettype == glsl_bool
        self.__padding = 4 - signature.rettype.num_rows

    def make_result_handler(self, invocation, output_var):
        statements = '  if({0})\n'.format(invocation)
        statements += '    {0} = vec4(1.0, 1.0, 0.0, 1.0);\n'.format(
            output_var)
        statements += '  else\n'
        statements += '    {0} = vec4(0.0, 0.0, 1.0, 1.0);\n'.format(
            output_var)
        return statements

    def convert_to_float(self, value):
        """Convert the given vector or scalar value to a list of
        floats representing the expected color produced by the test.
        """
        if value:
            return [1.0, 1.0, 0.0, 1.0]
        else:
            return [0.0, 0.0, 1.0, 1.0]

    def draw_test(self, test_vector, draw_command):
        return draw_command

    def result_vector(self, test_vector):
        return self.convert_to_float(test_vector.result)

    def testname_suffix(self):
        return '-using-if'


class IntComparator(Comparator):
    """Comparator that tests functions returning ints or ivecs using a
    strict equality test.

    This comparator causes code to be generated in the following form:

        rettype result = func(args);
        output_var = result == expected ? vec4(0.0, 1.0, 0.0, 1.0)
                                        : vec4(1.0, 0.0, 0.0, 1.0);
    """
    def __init__(self, signature):
        self.__signature = signature

    def make_additional_declarations(self):
        return 'uniform {0} expected;\n'.format(self.__signature.rettype)

    def make_result_handler(self, invocation, output_var):
        statements = '  {0} result = {1};\n'.format(
            self.__signature.rettype, invocation)
        statements += '  {v} = {cond} ? {green} : {red};\n'.format(
            v=output_var, cond='result == expected',
            green='vec4(0.0, 1.0, 0.0, 1.0)',
            red='vec4(1.0, 0.0, 0.0, 1.0)')
        return statements

    def draw_test(self, test_vector, draw_command):
        test = 'uniform {0} expected {1}\n'.format(
            shader_runner_type(self.__signature.rettype),
            shader_runner_format(column_major_values(test_vector.result)))
        test += draw_command
        return test

    def result_vector(self, test_vector):
        return [0.0, 1.0, 0.0, 1.0]


class FloatComparator(Comparator):
    """Comparator that tests functions returning floats or vecs using a
    strict equality test.

    This comparator causes code to be generated in the following form:

        rettype result = func(args);
        output_var = distance(result, expected) <= tolerance
                     ? vec4(0.0, 1.0, 0.0, 1.0) : vec4(1.0, 0.0, 0.0, 1.0);
    """
    def __init__(self, signature):
        self.__signature = signature

    def make_additional_declarations(self):
        decls = 'uniform float tolerance;\n'
        decls += 'uniform {0} expected;\n'.format(self.__signature.rettype)
        return decls

    def make_indexers(self):
        """Build a list of strings which index into every possible
        value of the result.  For example, if the result is a vec2,
        then build the indexers ['[0]', '[1]'].
        """
        if self.__signature.rettype.num_cols == 1:
            col_indexers = ['']
        else:
            col_indexers = ['[{0}]'.format(i)
                            for i in range(self.__signature.rettype.num_cols)]
        if self.__signature.rettype.num_rows == 1:
            row_indexers = ['']
        else:
            row_indexers = ['[{0}]'.format(i)
                            for i in range(self.__signature.rettype.num_rows)]
        return [col_indexer + row_indexer
                for col_indexer in col_indexers
                for row_indexer in row_indexers]

    def make_result_handler(self, invocation, output_var):
        statements = '  {0} result = {1};\n'.format(
            self.__signature.rettype, invocation)
        # Can't use distance when testing itself, or when the rettype
        # is a matrix.
        if self.__signature.name == 'distance' or \
                self.__signature.rettype.is_matrix:
            statements += '  {0} residual = result - expected;\n'.format(
                self.__signature.rettype)
            statements += '  float error_sq = {0};\n'.format(
                ' + '.join(
                    'residual{0} * residual{0}'.format(indexer)
                    for indexer in self.make_indexers()))
            condition = 'error_sq <= tolerance * tolerance'
        else:
            condition = 'distance(result, expected) <= tolerance'
        statements += '  {v} = {cond} ? {green} : {red};\n'.format(
            v=output_var, cond=condition, green='vec4(0.0, 1.0, 0.0, 1.0)',
            red='vec4(1.0, 0.0, 0.0, 1.0)')
        return statements

    def draw_test(self, test_vector, draw_command):
        test = 'uniform {0} expected {1}\n'.format(
            shader_runner_type(self.__signature.rettype),
            shader_runner_format(column_major_values(test_vector.result)))
        test += 'uniform float tolerance {0}\n'.format(
            shader_runner_format([test_vector.tolerance]))
        test += draw_command
        return test

    def result_vector(self, test_vector):
        return [0.0, 1.0, 0.0, 1.0]


class ShaderTest(object):
    """Class used to build a test of a single built-in.  This is an
    abstract base class--derived types should override test_prefix(),
    make_vertex_shader(), make_fragment_shader(), and other functions
    if necessary.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, signature, test_vectors, use_if):
        """Prepare to build a test for a single built-in.  signature
        is the signature of the built-in (a key from the
        builtin_function.test_suite dict), and test_vectors is the
        list of test vectors for testing the given builtin (the
        corresponding value from the builtin_function.test_suite
        dict).

        If use_if is True, then the generated test checks the result
        by using it in an if statement--this only works for builtins
        returning bool.
        """
        self._signature = signature
        self._test_vectors = test_vectors

        # Size of the rectangles drawn by the test.
        self.rect_width = 4
        self.rect_height = 4
        # shader_runner currently defaults to a 250x250 window.  We
        # could reduce window size to cut test time, but there are
        # platform-dependent limits we haven't really characterized
        # (smaller on Linux than Windows, but still present in some
        # window managers).
        self.win_width = 250
        self.win_height = 250
        self.tests_per_row = (self.win_width // self.rect_width)
        self.test_rows = (self.win_height // self.rect_height)

        if use_if:
            self._comparator = BoolIfComparator(signature)
        elif signature.rettype.base_type == glsl_bool:
            self._comparator = BoolComparator(signature)
        elif signature.rettype.base_type == glsl_float:
            self._comparator = FloatComparator(signature)
        elif signature.rettype.base_type in (glsl_int, glsl_uint):
            self._comparator = IntComparator(signature)
        elif signature.rettype.base_type in (glsl_int64_t, glsl_uint64_t):
            self._comparator = IntComparator(signature)
        else:
            raise Exception('Unexpected rettype {0}'.format(signature.rettype))

    def glsl_version(self):
        return self._signature.version_introduced

    def draw_command(self, test_num):
        x = (test_num % self.tests_per_row) * self.rect_width
        y = (test_num // self.tests_per_row) * self.rect_height
        assert(y + self.rect_height <= self.win_height)
        return 'draw rect ortho {0} {1} {2} {3}\n'.format(x, y,
                                                          self.rect_width,
                                                          self.rect_height)

    def probe_command(self, test_num, probe_vector):
        return 'probe rect rgba ({0}, {1}, {2}, {3}) ({4}, {5}, {6}, {7})\n'.format(
            (test_num % self.tests_per_row) * self.rect_width,
            (test_num // self.tests_per_row) * self.rect_height,
            self.rect_width,
            self.rect_height,
            probe_vector[0], probe_vector[1], probe_vector[2], probe_vector[3])

    def extensions(self):
        ext = []
        if self._signature.extension:
            ext.append(self._signature.extension)
        return ext

    def make_additional_requirements(self):
        """Return a string that should be included in the test's
        [require] section.
        """
        return ''

    @abc.abstractmethod
    def test_prefix(self):
        """Return the prefix that should be used in the test file name
        to identify the type of test, e.g. "vs" for a vertex shader
        test.
        """

    def make_vertex_shader(self):
        """Return the vertex shader for this test (or None if this
        test doesn't require a vertex shader).  No need to
        reimplement this function in classes that don't use vertex
        shaders.
        """
        return None

    def make_tess_ctrl_shader(self):
        """Return the tessellation control shader for this test
        (or None if this test doesn't require a geometry shader).
        No need to reimplement this function in classes that don't
        use geometry shaders.
        """
        return None

    def make_tess_eval_shader(self):
        """Return the tessellation evaluation shader for this test
        (or None if this test doesn't require a geometry shader).
        No need to reimplement this function in classes that don't
        use geometry shaders.
        """
        return None

    def make_geometry_shader(self):
        """Return the geometry shader for this test (or None if this
        test doesn't require a geometry shader).  No need to
        reimplement this function in classes that don't use geometry
        shaders.
        """
        return None

    def make_geometry_layout(self):
        """Return the geometry layout for this test (or None if this
        test doesn't require a geometry layout section).  No need to
        reimplement this function in classes that don't use geometry
        shaders.
        """
        return None

    def make_fragment_shader(self):
        """Return the fragment shader for this test (or None if this
        test doesn't require a fragment shader).  No need to
        reimplement this function in classes that don't use fragment
        shaders.
        """
        return None

    def make_compute_shader(self):
        """Return the compute shader for this test (or None if this test
        doesn't require a compute shader).  No need to reimplement
        this function in classes that don't use compute shaders.
        """
        return None

    def needs_probe_per_draw(self):
        """Returns whether the test needs the probe to be immediately after each

        draw call.
        """
        return False

    def make_test_shader(self, additional_declarations, prefix_statements,
                         output_var, suffix_statements):
        """Generate the shader code necessary to test the built-in.
        additional_declarations is a string containing any
        declarations that need to be before the main() function of the
        shader.  prefix_statements is a string containing any
        additional statements than need to be inside the main()
        function of the shader, before the built-in function is
        called.  output_var is the variable that the result of the
        built-in function should be assigned to, after conversion to a
        vec4.  suffix_statements is a string containing any additional
        statements that need to be inside the main() function of the
        shader, after the built-in function is called.
        """
        shader = ''
        for ext in self.extensions():
            shader += '#extension GL_{0} : require\n'.format(ext)
        shader += additional_declarations
        for i in range(len(self._signature.argtypes)):
            shader += 'uniform {0} arg{1};\n'.format(
                self._signature.argtypes[i], i)
        shader += self._comparator.make_additional_declarations()
        shader += '\n'
        shader += 'void main()\n'
        shader += '{\n'
        shader += prefix_statements
        invocation = self._signature.template.format(
            *['arg{0}'.format(i)
              for i in range(len(self._signature.argtypes))])
        shader += self._comparator.make_result_handler(invocation, output_var)
        shader += suffix_statements
        shader += '}\n'
        return shader

    def make_test_init(self):
        """Generate initialization for the test.
        """
        return ''

    def make_test(self):
        """Make the complete shader_runner test file, and return it as
        a string.
        """
        test = self.make_test_init()
        for test_num, test_vector in enumerate(self._test_vectors):
            for i in range(len(test_vector.arguments)):
                test += 'uniform {0} arg{1} {2}\n'.format(
                    shader_runner_type(self._signature.argtypes[i]),
                    i, shader_runner_format(
                        column_major_values(test_vector.arguments[i])))
            test += self._comparator.draw_test(test_vector,
                                               self.draw_command(test_num))
            if self.needs_probe_per_draw():
                result_color = self._comparator.result_vector(test_vector)
                test += self.probe_command(test_num, result_color)

        if not self.needs_probe_per_draw():
            for test_num, test_vector in enumerate(self._test_vectors):
                result_color = self._comparator.result_vector(test_vector)
                test += self.probe_command(test_num, result_color)
        return test

    def filename(self):
        argtype_names = '-'.join(
            str(argtype) for argtype in self._signature.argtypes)
        if self.extensions():
            subdir = self.extensions()[0].lower()
        else:
            subdir = 'glsl-{0:1.2f}'.format(float(self.glsl_version()) / 100)
        return os.path.join(
            'spec', subdir, 'execution', 'built-in-functions',
            '{0}-{1}-{2}{3}.shader_test'.format(
                self.test_prefix(), self._signature.name, argtype_names,
                self._comparator.testname_suffix()))

    def generate_shader_test(self):
        """Generate the test and write it to the output file."""
        shader_test = '[require]\n'
        shader_test += 'GLSL >= {0:1.2f}\n'.format(
            float(self.glsl_version()) / 100)
        for extension in self.extensions():
            shader_test += 'GL_{}\n'.format(extension)
        shader_test += self.make_additional_requirements()
        shader_test += '\n'
        vs = self.make_vertex_shader()
        if vs:
            shader_test += '[vertex shader]\n'
            shader_test += vs
            shader_test += '\n'
        tcs = self.make_tess_ctrl_shader()
        if tcs:
            shader_test += '[tessellation control shader]\n'
            shader_test += tcs
            shader_test += '\n'
        tes = self.make_tess_eval_shader()
        if tes:
            shader_test += '[tessellation evaluation shader]\n'
            shader_test += tes
            shader_test += '\n'
        gs = self.make_geometry_shader()
        if gs:
            shader_test += '[geometry shader]\n'
            shader_test += gs
            shader_test += '\n'
        gl = self.make_geometry_layout()
        if gl:
            shader_test += '[geometry layout]\n'
            shader_test += gl
            shader_test += '\n'
        fs = self.make_fragment_shader()
        if fs:
            shader_test += '[fragment shader]\n'
            shader_test += fs
            shader_test += '\n'
        cs = self.make_compute_shader()
        if cs:
            shader_test += '[compute shader]\n'
            shader_test += cs
            shader_test += '\n'
        shader_test += '[test]\n'
        shader_test += 'clear color 0.0 0.0 1.0 0.0\n'
        shader_test += 'clear\n'
        shader_test += self.make_test()
        filename = self.filename()
        dirname = os.path.dirname(filename)
        utils.safe_makedirs(dirname)
        with open(filename, 'w') as f:
            f.write(shader_test)


class VertexShaderTest(ShaderTest):
    """Derived class for tests that exercise the built-in in a vertex
    shader.
    """
    def test_prefix(self):
        return 'vs'

    def make_vertex_shader(self):
        if self.glsl_version() >= 140:
            return self.make_test_shader(
                'in vec4 piglit_vertex;\n' +
                'out vec4 color;\n',
                '  gl_Position = piglit_vertex;\n',
                'color', '')
        else:
            return self.make_test_shader(
                'varying vec4 color;\n',
                '  gl_Position = gl_Vertex;\n',
                'color', '')

    def make_fragment_shader(self):
        shader = '''varying vec4 color;

void main()
{
  gl_FragColor = color;
}
'''
        return shader


class TessellationShaderTest(ShaderTest):
    """Abstract class for tests that exercise the built-in in
    tessellation shaders.
    """

    def glsl_version(self):
        return max(150, ShaderTest.glsl_version(self))

    def make_additional_requirements(self):
        return 'GL_ARB_tessellation_shader'

    def extensions(self):
        ext = []
        if self._signature.extension:
            ext.append(self._signature.extension)
        ext.append("ARB_tessellation_shader")
        return ext

    def draw_command(self, test_num):
        x = (test_num % self.tests_per_row) * self.rect_width
        y = (test_num // self.tests_per_row) * self.rect_height
        assert(y + self.rect_height <= self.win_height)
        return 'draw rect ortho patch {0} {1} {2} {3}\n'.format(x, y,
                                                                self.rect_width,
                                                                self.rect_height)

    def make_vertex_shader(self):
        shader = \
"""in vec4 piglit_vertex;
out vec4 vertex_to_tcs;

void main()
{
     vertex_to_tcs = piglit_vertex;
}
"""
        return shader

    def make_fragment_shader(self):
        shader = \
"""in vec4 color_to_fs;

void main()
{
  gl_FragColor = color_to_fs;
}
"""
        return shader


class TessCtrlShaderTest(TessellationShaderTest):
    """Derived class for tests that exercise the built-in in a
    tessellation control shader.
    """

    def test_prefix(self):
        return 'tcs'

    def make_tess_ctrl_shader(self):
        additional_declarations = \
"""layout(vertices = 4) out;
in vec4 vertex_to_tcs[];
out vec4 vertex_to_tes[];
patch out vec4 color_to_tes;
"""
        body = \
"""  vertex_to_tes[gl_InvocationID] = vertex_to_tcs[gl_InvocationID];
  color_to_tes = tmp_color;
  gl_TessLevelOuter = float[4](1.0, 1.0, 1.0, 1.0);
  gl_TessLevelInner = float[2](1.0, 1.0);
"""
        shader = self.make_test_shader(
            additional_declarations,
            '  vec4 tmp_color;\n',
            'tmp_color',
            body)
        return shader

    def make_tess_eval_shader(self):
        shader = \
"""#extension GL_ARB_tessellation_shader : require
layout(quads) in;

in vec4 vertex_to_tes[];
patch in vec4 color_to_tes;
out vec4 color_to_fs;
void main() {
  gl_Position = mix(mix(vertex_to_tes[0], vertex_to_tes[1], gl_TessCoord.x),
                    mix(vertex_to_tes[2], vertex_to_tes[3], gl_TessCoord.x),
                    gl_TessCoord.y);
  color_to_fs = color_to_tes;
}
"""
        return shader


class GeometryShaderTest(ShaderTest):
    """Derived class for tests that exercise the built-in in a
    geometry shader.
    """
    def test_prefix(self):
        return 'gs'

    def glsl_version(self):
        return max(150, ShaderTest.glsl_version(self))

    def make_vertex_shader(self):
        shader = ''
        shader += "in vec4 piglit_vertex;\n"
        shader += "out vec4 vertex_to_gs;\n"

        shader += "void main()\n"
        shader += "{\n"
        shader += "     vertex_to_gs = piglit_vertex;\n"
        shader += "}\n"

        return shader

    def make_geometry_shader(self):
        additional_declarations = ''
        additional_declarations += 'layout(triangles) in;\n'
        additional_declarations \
            += 'layout(triangle_strip, max_vertices = 3) out;\n'
        additional_declarations += 'in vec4 vertex_to_gs[3];\n'
        additional_declarations += 'out vec4 color;\n'
        return self.make_test_shader(
            additional_declarations,
            '  vec4 tmp_color;\n',
            'tmp_color',
            '  for (int i = 0; i < 3; i++) {\n'
            '    gl_Position = vertex_to_gs[i];\n'
            '    color = tmp_color;\n'
            '    EmitVertex();\n'
            '  }\n')

    def make_fragment_shader(self):
        shader = '''varying vec4 color;

void main()
{
  gl_FragColor = color;
}
'''
        return shader


class FragmentShaderTest(ShaderTest):
    """Derived class for tests that exercise the built-in in a
    fragment shader.
    """
    def test_prefix(self):
        return 'fs'

    def make_vertex_shader(self):
        shader = ""
        if self.glsl_version() >= 140:
            shader += "in vec4 piglit_vertex;\n"

        shader += "void main()\n"
        shader += "{\n"
        if self.glsl_version() >= 140:
            shader += "        gl_Position = piglit_vertex;\n"
        else:
            shader += "        gl_Position = gl_Vertex;\n"
        shader += "}\n"

        return shader

    def make_fragment_shader(self):
        return self.make_test_shader('', '', 'gl_FragColor', '')


class ComputeShaderTest(ShaderTest):
    """Derived class for tests that exercise the built-in in a
    compute shader.
    """
    def test_prefix(self):
        return 'cs'

    def glsl_version(self):
        return max(430, ShaderTest.glsl_version(self))

    def make_compute_shader(self):
        additional_declarations = 'writeonly uniform image2D tex;\n'
        num_tests = len(self._test_vectors)
        layout_tmpl = 'layout(local_size_x = {0}) in;\n'
        additional_declarations += layout_tmpl.format(num_tests)
        return self.make_test_shader(
            additional_declarations,
            '  vec4 tmp_color;\n',
            'tmp_color',
            '  ivec2 coord = ivec2(gl_GlobalInvocationID.xy);\n'
            '  imageStore(tex, coord, tmp_color);\n')

    def make_test_init(self):
        return '''uniform int tex 0
texture rgbw 0 ({0}, 1) GL_RGBA8
image texture 0 GL_RGBA8
fb tex 2d 0
'''.format(len(self._test_vectors))


    def draw_command(self, test_num):
        return 'compute 1 1 1\n'

    def probe_command(self, test_num, probe_vector):
        # Note: shader_runner uses a 250x250 window so we must
        # ensure that test_num <= 250.
        return 'probe rgb {0} 0 {1} {2} {3} {4}\n'.format(test_num % 250,
                                                          probe_vector[0],
                                                          probe_vector[1],
                                                          probe_vector[2],
                                                          probe_vector[3])
    def needs_probe_per_draw(self):
        return True

def all_tests():
    for use_if in [False, True]:
        for signature, test_vectors in sorted(test_suite.items()):
            if use_if and signature.rettype != glsl_bool:
                continue
            yield VertexShaderTest(signature, test_vectors, use_if)
            yield TessCtrlShaderTest(signature, test_vectors, use_if)
            yield GeometryShaderTest(signature, test_vectors, use_if)
            yield FragmentShaderTest(signature, test_vectors, use_if)
            yield ComputeShaderTest(signature, test_vectors, use_if)


def main():
    desc = 'Generate shader tests that test built-in functions using uniforms'
    usage = 'usage: %prog [-h] [--names-only]'
    parser = optparse.OptionParser(description=desc, usage=usage)
    parser.add_option(
        '--names-only',
        dest='names_only',
        action='store_true',
        help="Don't output files, just generate a list of filenames to stdout")
    options, args = parser.parse_args()
    for test in all_tests():
        if not options.names_only:
            test.generate_shader_test()
        print(test.filename())


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2011 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# Generate a pair of glsl parser tests for every overloaded version of
# every built-in function, which test that the built-in functions are
# handled properly when applied to constant arguments inside an array
# size declaration.
#
# In each pair of generated tests, one test exercises the built-in
# function in vertex shaders, and the other exercises it in fragment
# shaders.
#
# This program outputs, to stdout, the name of each file it generates.
# With the optional argument --names-only, it only outputs the names
# of the files; it doesn't generate them.

from __future__ import print_function, division, absolute_import
from builtin_function import *
import abc
import optparse
import os
import os.path

from six.moves import range

from modules import utils

class ParserTest(object):
    """Class used to build a test of a single built-in.  This is an
    abstract base class--derived types should override test_suffix(),
    output_var(), and other functions if necessary.
    """

    def __init__(self, signature, test_vectors):
        """Prepare to build a test for a single built-in.  signature
        is the signature of the built-in (a key from the
        builtin_function.test_suite dict), and test_vectors is the
        list of test vectors for testing the given builtin (the
        corresponding value from the builtin_function.test_suite
        dict).
        """
        self.__signature = signature
        self.__test_vectors = test_vectors

    def glsl_version(self):
        if self.__signature.version_introduced < 120:
            # Before version 1.20, built-in function invocations
            # weren't allowed in constant expressions.  So even if
            # this built-in was introduced prior to 1.20, test it in
            # version 1.20.
            return 120
        else:
            return self.__signature.version_introduced

    def version_directive(self):
        return '#version {0}\n'.format(self.glsl_version())

    def additional_declarations(self):
        """Return a string containing any additional declarations that
        should be placed after the version directive.  Returns the
        empty string by default.
        """
        return ''

    def additional_extensions(self):
        """Return a list (or other iterable) containing any additional
        extension requirements that the test has.  Returns the empty
        list by default.
        """
        ret = []
        if self.__signature.extension:
            ret.append(self.__signature.extension)

        return ret

    @abc.abstractmethod
    def test_suffix(self):
        """Return the suffix that should be used in the test file name
        to identify the type of shader, e.g. "vert" for a vertex
        shader test.
        """

    def make_condition(self, test_vector):
        """Generate a GLSL constant expression that should evaluate to
        true if the GLSL compiler's constant evaluation produces the
        correct result for the given test vector, and false if not.
        """
        invocation = self.__signature.template.format(
            *[glsl_constant(x) for x in test_vector.arguments])
        if self.__signature.rettype.base_type == glsl_float:
            # Test floating-point values within tolerance
            if self.__signature.name == 'distance':
                # Don't use the distance() function to test itself.
                return '{0} <= {1} && {1} <= {2}'.format(
                    test_vector.result - test_vector.tolerance,
                    invocation,
                    test_vector.result + test_vector.tolerance)
            elif self.__signature.rettype.is_matrix:
                # We can't apply distance() to matrices.  So apply it
                # to each column and root-sum-square the results.  It
                # is safe to use pow() here because its behavior is
                # verified in the pow() tests.
                terms = []
                for col in range(self.__signature.rettype.num_cols):
                    terms.append('pow(distance({0}[{1}], {2}), 2)'.format(
                        invocation, col,
                        glsl_constant(test_vector.result[:, col])))
                rss_distance = ' + '.join(terms)
                sq_tolerance = test_vector.tolerance * test_vector.tolerance
                return '{0} <= {1}'.format(
                    rss_distance, glsl_constant(sq_tolerance))
            else:
                return 'distance({0}, {1}) <= {2}'.format(
                    invocation, glsl_constant(test_vector.result),
                    glsl_constant(test_vector.tolerance))
        else:
            # Test non-floating point values exactly
            assert not self.__signature.rettype.is_matrix
            if self.__signature.name == 'equal':
                # Don't use the equal() function to test itself.
                assert self.__signature.rettype.is_vector
                terms = []
                for row in range(self.__signature.rettype.num_rows):
                    terms.append('{0}[{1}] == {2}'.format(
                        invocation, row,
                        glsl_constant(test_vector.result[row])))
                return ' && '.join(terms)
            elif self.__signature.rettype.is_vector:
                return 'all(equal({0}, {1}))'.format(
                    invocation, glsl_constant(test_vector.result))
            else:
                return '{0} == {1}'.format(
                    invocation, glsl_constant(test_vector.result))

    def make_shader(self):
        """Generate the shader code necessary to test the built-in."""
        shader = self.version_directive()
        if self.__signature.extension:
            shader += '#extension GL_{0} : require\n'.format(self.__signature.extension)
        shader += self.additional_declarations()
        shader += '\n'
        shader += 'void main()\n'
        shader += '{\n'
        lengths = []
        for i, test_vector in enumerate(self.__test_vectors):
            shader += '  float[{0} ? 1 : -1] array{1};\n'.format(
                self.make_condition(test_vector), i)
            lengths.append('array{0}.length()'.format(i))
        shader += '  {0} = vec4({1});\n'.format(
            self.output_var(), ' + '.join(lengths))
        shader += '}\n'
        return shader

    def filename(self):
        argtype_names = '-'.join(
            str(argtype) for argtype in self.__signature.argtypes)
        if self.__signature.extension:
            subdir = self.__signature.extension.lower()
        else:
            subdir = 'glsl-{0:1.2f}'.format(float(self.glsl_version()) / 100)
        return os.path.join(
            'spec', subdir, 'compiler', 'built-in-functions',
            '{0}-{1}.{2}'.format(
                self.__signature.name, argtype_names, self.test_suffix()))

    def generate_parser_test(self):
        """Generate the test and write it to the output file."""
        parser_test = '/* [config]\n'
        parser_test += ' * expect_result: pass\n'
        parser_test += ' * glsl_version: {0:1.2f}\n'.format(
            float(self.glsl_version()) / 100)
        req_extensions = list(self.additional_extensions())
        if req_extensions:
            parser_test += ' * require_extensions: {}\n'.format(
                ' '.join('GL_{}'.format(r) for r in req_extensions))
        parser_test += ' * [end config]\n'
        parser_test += ' *\n'
        parser_test += ' * Check that the following test vectors are constant'\
                       ' folded correctly:\n'
        for test_vector in self.__test_vectors:
            parser_test += ' * {0} => {1}\n'.format(
                self.__signature.template.format(
                    *[glsl_constant(arg) for arg in test_vector.arguments]),
                glsl_constant(test_vector.result))
        parser_test += ' */\n'
        parser_test += self.make_shader()
        filename = self.filename()
        dirname = os.path.dirname(filename)
        utils.safe_makedirs(dirname)
        with open(filename, 'w') as f:
            f.write(parser_test)


class VertexParserTest(ParserTest):
    """Derived class for tests that exercise the built-in in a vertex
    shader.
    """
    def test_suffix(self):
        return 'vert'

    def output_var(self):
        return 'gl_Position'


class GeometryParserTest(ParserTest):
    """Derived class for tests that exercise the built-in in a geometry
    shader.
    """
    def glsl_version(self):
        return max(150, ParserTest.glsl_version(self))

    def test_suffix(self):
        return 'geom'

    def output_var(self):
        return 'gl_Position'


class FragmentParserTest(ParserTest):
    """Derived class for tests that exercise the built-in in a fagment
    shader.
    """
    def test_suffix(self):
        return 'frag'

    def output_var(self):
        return 'gl_FragColor'


def all_tests():
    for signature, test_vectors in sorted(test_suite.items()):
        # Assignment operators other than = cannot be used in the constant
        # array size tests
        if not signature.name.startswith('op-assign'):
            yield VertexParserTest(signature, test_vectors)
            yield GeometryParserTest(signature, test_vectors)
            yield FragmentParserTest(signature, test_vectors)


def main():
    desc = 'Generate shader tests that test built-in functions using constant'\
           'array sizes'
    usage = 'usage: %prog [-h] [--names-only]'
    parser = optparse.OptionParser(description=desc, usage=usage)
    parser.add_option('--names-only',
                      dest='names_only',
                      action='store_true',
                      help="Don't output files, just generate a list of"
                           "filenames to stdout")
    options, args = parser.parse_args()

    for test in all_tests():
        if not options.names_only:
            test.generate_parser_test()
        print(test.filename())


if __name__ == '__main__':
    main()



# encoding=utf-8
# Copyright (c) 2014-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the results module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pytest
import six

from framework import exceptions
from framework import grouptools
from framework import results
from framework import status

from .backends import shared

# pylint: disable=no-self-use


class TestSubtests(object):
    """Tests for the Subtest class."""

    @pytest.fixture
    def subtest(self):
        return results.Subtests()

    def test_convert_statuses(self, subtest):
        """results.Subtests.__setitem__: converts strings to statues."""
        subtest['foo'] = 'pass'
        assert subtest['foo'] is status.PASS

    def test_to_json(self, subtest):
        """results.Subtests.to_json: sets values properly."""
        baseline = {
            'foo': status.PASS,
            'bar': status.CRASH,
            '__type__': 'Subtests',
        }

        subtest['foo'] = status.PASS
        subtest['bar'] = status.CRASH

        assert baseline == subtest.to_json()

    def test_from_dict(self, subtest):
        """results.Subtests.from_dict: restores values properly"""
        subtest['foo'] = status.PASS
        subtest['bar'] = status.CRASH

        test = results.Subtests.from_dict(subtest.to_json())

        # Subtests don't have equality methods, so convert them to dicts before
        # comparing.
        assert dict(subtest) == dict(test)

    def test_from_dict_instance(self, subtest):
        """results.Subtests.from_dict: restores values properly"""
        subtest['foo'] = status.PASS
        test = results.Subtests.from_dict(subtest.to_json())

        assert test['foo'] is status.PASS


class TestTestResult(object):
    """Tests for the TestResult class."""

    class TestFromDict(object):
        """Tests for TestResult.from_dict."""

        def test_inst(self):
            """results.TestResult.from_dict: returns a TestResult."""
            test = results.TestResult.from_dict({'result': 'pass'})
            assert isinstance(test, results.TestResult)

        class TestAttributes(object):
            """Tests attribute restoration for the from_dict method."""

            @classmethod
            def setup_class(cls):
                """Setup state for all tests."""
                cls.dict = {
                    'returncode': 100,
                    'err': 'this is an err',
                    'out': 'this is some text',
                    'time': {
                        'start': 0.5,
                        'end': 0.9,
                    },
                    'environment': 'some env stuff',
                    'subtests': {
                        'a': 'pass',
                        'b': 'fail',
                    },
                    'result': 'crash',
                    'exception': 'an exception',
                    'dmesg': 'this is dmesg',
                    'pid': [1934],
                }

                cls.test = results.TestResult.from_dict(cls.dict)

            def test_returncode(self):
                """sets returncode properly."""
                assert self.test.returncode == self.dict['returncode']

            def test_err(self):
                """sets err properly."""
                assert self.test.err == self.dict['err']

            def test_out(self):
                """sets out properly."""
                assert self.test.out == self.dict['out']

            def test_time(self):
                """sets time properly.

                Ultimatley time needs to be converted to a TimeAttribute object,
                not a dictionary, however, that functionality is handled by
                backends.json.piglit_decoder, not by the from_dict method of
                TestResult. So in this case the test is that the object is a
                dictionary, not a TimeAttribute because this method shouldn't
                make the coversion.

                """
                # pylint: disable=unsubscriptable-object
                assert self.test.time.start == self.dict['time']['start']
                assert self.test.time.end == self.dict['time']['end']

            def test_environment(self):
                """sets environment properly."""
                assert self.test.environment == self.dict['environment']

            def test_exception(self):
                """sets exception properly."""
                assert self.test.exception == self.dict['exception']

            def test_subtests(self):
                """sets subtests properly."""
                assert self.test.subtests == self.dict['subtests']

            def test_subtests_type(self):
                """subtests are Status instances."""
                assert self.test.subtests['a'] is status.PASS
                assert self.test.subtests['b'] is status.FAIL

            def test_dmesg(self):
                """sets dmesg properly."""
                assert self.test.dmesg == self.dict['dmesg']

            def test_pid(self):
                """sets pid properly."""
                assert self.test.pid == self.dict['pid']

        class TestResult(object):
            """Tests for TestResult.result getter and setter methods."""

            @pytest.fixture
            def result(self):
                return results.TestResult('pass')

            def test_getter_no_subtests(self, result):
                """Getter returns the result when there are no subtests."""
                assert result.result is status.PASS

            def test_getter_with_subtests(self, result):
                """Getter returns worst subtest when subtests are present."""
                result.subtests['a'] = status.PASS
                result.subtests['b'] = status.CRASH

                assert result.result is status.CRASH

            def test_getter_subtests_crash(self, result):
                """Doesn't mask crashes.

                Crash is somewhat unique in that subtests can run and pass, but
                then the test can crash before all subtests have been run.
                """
                result.result = status.CRASH
                result.subtests['a'] = status.PASS
                result.subtests['b'] = status.SKIP

                assert result.result is status.CRASH

            def test_setter(self, result):
                """setter makes the result a status."""
                result.result = status.FAIL
                assert result.result is status.FAIL

            def test_setter_invalid(self, result):
                """setter raises PiglitFatalError for invalid values."""
                with pytest.raises(exceptions.PiglitFatalError):
                    result.result = 'foo'

    class TestToJson(object):
        """Tests for the attributes of the to_json method."""

        @classmethod
        def setup_class(cls):
            """Setup state for all tests."""
            test = results.TestResult()
            test.returncode = 100
            test.err = 'this is an err'
            test.out = 'this is some text'
            test.time.start = 0.3
            test.time.end = 0.5
            test.environment = 'some env stuff'
            test.subtests.update({
                'a': 'pass',
                'b': 'fail'})
            test.result = 'crash'
            test.exception = 'an exception'
            test.dmesg = 'this is dmesg'
            test.pid = 1934
            test.traceback = 'a traceback'

            cls.test = test
            cls.json = test.to_json()

        def test_returncode(self):
            """results.TestResult.to_json: sets the returncode correctly"""
            assert self.test.returncode == self.json['returncode']

        def test_err(self):
            """results.TestResult.to_json: sets the err correctly"""
            assert self.test.err == self.json['err']

        def test_out(self):
            """results.TestResult.to_json: sets the out correctly"""
            assert self.test.out == self.json['out']

        def test_exception(self):
            """results.TestResult.to_json: sets the exception correctly"""
            assert self.test.exception == self.json['exception']

        def test_time(self):
            """results.TestResult.to_json: sets the time correctly"""
            # pylint: disable=unsubscriptable-object
            assert self.test.time.start == self.json['time']['start']
            assert self.test.time.end == self.json['time']['end']

        def test_environment(self):
            """results.TestResult.to_json: sets the environment correctly"""
            assert self.test.environment == self.json['environment']

        def test_subtests(self):
            """results.TestResult.to_json: sets the subtests correctly"""
            assert self.test.subtests['a'] == self.json['subtests']['a']
            assert self.test.subtests['b'] == self.json['subtests']['b']
            assert self.json['subtests']['__type__']

        def test_type(self):
            """results.TestResult.to_json: adds the __type__ hint"""
            assert self.json['__type__'] == 'TestResult'

        def test_dmesg(self):
            """results.TestResult.to_json: Adds the dmesg attribute"""
            assert self.test.dmesg == self.json['dmesg']

        def test_pid(self):
            """results.TestResult.to_json: Adds the pid attribute"""
            assert self.test.pid == self.json['pid']

        def test_traceback(self):
            """results.TestResult.to_json: Adds the traceback attribute"""
            assert self.test.traceback == self.json['traceback']

    class TestUpdate(object):
        """Tests for TestResult.update."""

        def test_no_subtests(self):
            """results.TestResult.update: result is updated"""
            test = results.TestResult('pass')
            test.update({'result': 'incomplete'})
            assert test.result == 'incomplete'

        def test_with_subtests(self):
            """results.TestResult.update: subests are updated"""
            test = results.TestResult('pass')
            test.update({'subtest': {'result': 'incomplete'}})
            assert test.subtests['result'] == 'incomplete'

    class TestTotals(object):
        """Test the totals generated by TestrunResult.calculate_group_totals().
        """
        @classmethod
        def setup_class(cls):
            """setup state for all tests."""
            pass_ = results.TestResult('pass')
            fail = results.TestResult('fail')
            crash = results.TestResult('crash')
            skip = results.TestResult('skip')
            tr = results.TestrunResult()
            tr.tests = {
                'oink': pass_,
                grouptools.join('foo', 'bar'): fail,
                grouptools.join('foo', 'foo', 'bar'): crash,
                grouptools.join('foo', 'foo', 'oink'): skip,
            }

            tr.calculate_group_totals()
            cls.test = tr.totals

        def test_root(self):
            """The root is correct."""
            root = results.Totals()
            root['pass'] += 1
            root['fail'] += 1
            root['crash'] += 1
            root['skip'] += 1

            assert dict(self.test['root']) == dict(root)

        def test_recurse(self):
            """Recurses correctly."""
            expected = results.Totals()
            expected['fail'] += 1
            expected['crash'] += 1
            expected['skip'] += 1
            assert dict(self.test['foo']) == dict(expected)

        def test_two_parents(self):
            """Handles multiple parents correctly."""
            expected = results.Totals()
            expected['crash'] += 1
            expected['skip'] += 1
            assert dict(self.test[grouptools.join('foo', 'foo')]) == \
                dict(expected)

    class TestTotalsWithSubtests(object):
        """results.TestrunResult.totals: Tests with subtests are handled
        correctly.
        """

        @classmethod
        def setup_class(cls):
            """Setup all tests."""
            tr = results.TestResult('crash')
            tr.subtests['foo'] = status.PASS
            tr.subtests['bar'] = status.CRASH
            tr.subtests['oink'] = status.FAIL

            run = results.TestrunResult()
            run.tests[grouptools.join('sub', 'test')] = tr
            run.calculate_group_totals()

            cls.test = run.totals

        def test_root(self):
            """The root is correct with subtests."""
            expect = results.Totals()
            expect['pass'] += 1
            expect['crash'] += 1
            expect['fail'] += 1
            assert dict(self.test['root']) == dict(expect)

        def test_node(self):
            """Tests with subtests are treated as groups."""
            key = grouptools.join('sub', 'test')
            assert key in self.test

        def test_node_values(self):
            """Tests with subtests values are correct."""
            expect = results.Totals()
            expect['pass'] += 1
            expect['crash'] += 1
            expect['fail'] += 1
            assert dict(self.test[grouptools.join('sub', 'test')]) == \
                dict(expect)


class TestStringDescriptor(object):
    """Test class for StringDescriptor."""

    @pytest.fixture
    def test(self):
        class Test(object):
            val = results.StringDescriptor('test')

        return Test()

    def test_get_default(self, test):
        """results.StringDescriptor.__get__: returns default when unset"""
        assert test.val == ''

    def test_set_no_replace(self, test):
        """results.StringDescriptor.__set__: instance is not replaced

        This works by setting the value to a valid value (either bytes or str)
        and then trying to set it to an invalid value (int). If the assignment
        succeeds then the instance has been replaced, if not, it hasn't.

        """
        test.val = 'foo'
        with pytest.raises(TypeError):
            test.val = 1  # pylint: disable=redefined-variable-type

    def test_set_str(self, test):
        """results.StringDescriptor.__set__: str is stored directly"""
        inst = 'foo'
        test.val = inst
        assert test.val == inst

    def test_set_bytes(self, test):
        """results.StringDescriptor.__set__: converts bytes to str"""
        inst = b'foo'
        test.val = inst
        assert test.val == 'foo'

    def test_set_str_unicode_literals(self, test):
        """results.StringDescriptor.__set__: handles unicode litterals in strs.
        """
        test.val = '\ufffd'
        assert test.val == '�'

    def test_delete(self, test):
        """results.StringDescriptor.__delete__: raises NotImplementedError"""
        with pytest.raises(NotImplementedError):
            del test.val


class TestTotals(object):
    """Tests for the totals class."""

    def test_totals_false(self):
        """bool() returns False when all values are 0."""
        assert not bool(results.Totals())

    # The tuple is required because of the timeout status, which conflicts with
    # the timeout pylint plugin
    @pytest.mark.parametrize(
        "key", ((x, ) for x in six.iterkeys(results.Totals())))
    def test_totals_true(self, key):
        """bool() returns True when any value is not 0."""
        test = results.Totals()
        test[key[0]] += 1
        assert bool(test)


class TestTestrunResult(object):
    """Tests for the TestrunResult class."""

    class TestToJson(object):
        """Test TestrunResult.to_json method."""

        @classmethod
        def setup_class(cls):
            """Setup values used by all tests."""
            test = results.TestrunResult()
            test.info = {
                'system': {
                    'uname': 'this is uname',
                    'glxinfo': 'glxinfo',
                    'clinfo': 'clinfo',
                    'wglinfo': 'wglinfo',
                    'lspci': 'this is lspci',
                }
            }
            test.name = 'name'
            test.options = {'some': 'option'}
            test.time_elapsed.end = 1.23
            test.tests = {'a test': results.TestResult('pass')}

            cls.test = test.to_json()

        def test_name(self):
            """name is properly encoded."""
            assert self.test['name'] == 'name'

        def test_info(self):
            assert self.test['info'] == {
                'system': {
                    'uname': 'this is uname',
                    'glxinfo': 'glxinfo',
                    'clinfo': 'clinfo',
                    'wglinfo': 'wglinfo',
                    'lspci': 'this is lspci',
                }
            }

        def test_options(self):
            """options is properly encoded."""
            assert dict(self.test['options']) == {'some': 'option'}

        def test_time(self):
            """time_elapsed is properly encoded."""
            assert self.test['time_elapsed'].end == 1.23

        def test_tests(self):
            """tests is properly encoded."""
            assert self.test['tests']['a test']['result'] == 'pass'

        def test_type(self):
            """__type__ is added."""
            assert self.test['__type__'] == 'TestrunResult'

    class TestFromDict(object):
        """Tests for TestrunResult.from_dict."""

        @pytest.fixture(scope="module")
        def inst(self):
            return results.TestrunResult.from_dict(shared.JSON)

        @pytest.mark.parametrize("attrib", [
            'name', 'results_version', 'info', 'options',
        ])
        def test_attribs_restored(self, attrib, inst):
            """tests for basic attributes."""
            assert shared.JSON[attrib] == getattr(inst, attrib)

        def test_tests(self, inst):
            """tests is restored correctly."""
            assert inst.tests.keys() == shared.JSON['tests'].keys()

        def test_test_type(self, inst):
            """tests is restored correctly."""
            assert isinstance(
                inst.tests['spec@!opengl 1.0@gl-1.0-readpixsanity'],
                results.TestResult)

        def test_totals(self, inst):
            """totals is restored correctly."""
            baseline = shared.JSON['totals'].copy()
            for s in six.itervalues(baseline):
                del s['__type__']
            assert baseline == dict(inst.totals)

        def test_time_elapsed(self, inst):
            """time_elapsed is restored correctly."""
            assert inst.time_elapsed.start == \
                shared.JSON['time_elapsed']['start']
            assert inst.time_elapsed.end == \
                shared.JSON['time_elapsed']['end']

    class TestGetResult(object):
        """Tests for TestrunResult.get_result."""

        @classmethod
        def setup_class(cls):
            """setup state for all tests."""
            tr = results.TestResult('crash')
            tr.subtests['foo'] = status.PASS

            run = results.TestrunResult()
            run.tests['sub'] = tr
            run.tests['test'] = results.TestResult('pass')
            run.calculate_group_totals()

            cls.inst = run

        def test_get_test(self):
            """gets non-subtests."""
            assert self.inst.get_result('test') == 'pass'

        def test_get_subtest(self):
            """gets subtests."""
            assert self.inst.get_result(grouptools.join('sub', 'foo')) == 'pass'

        def test_get_nonexist(self):
            """raises KeyError if test doesn't exist."""
            with pytest.raises(KeyError):
                self.inst.get_result('fooobar')


class TestTimeAttribute(object):
    """Tests for the TimeAttribute class."""

    def test_to_json(self):
        """returns expected dictionary."""
        baseline = {'start': 0.1, 'end': 1.0}
        test = results.TimeAttribute(**baseline)
        baseline['__type__'] = 'TimeAttribute'

        assert baseline == test.to_json()

    def test_from_dict(self):
        """returns expected value."""
        # Type is included because to_json() adds it.
        baseline = {'start': 0.1, 'end': 1.0, '__type__': 'TimeAttribute'}
        test = results.TimeAttribute.from_dict(baseline).to_json()

        assert baseline == test

    def test_total(self):
        """returns the difference between end and start."""
        test = results.TimeAttribute(1.0, 5.0)
        assert test.total == 4.0

    def test_delta(self):
        """results.TimeAttribute.delta: returns the delta of the values"""
        test = results.TimeAttribute(1.0, 5.0)
        assert test.delta == '0:00:04'



# coding=utf-8
# Copyright (c) 2013-2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Test integreation for the X Test Suite """

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os
import re
import subprocess
import itertools

from framework import grouptools, exceptions, core
from framework.profile import TestProfile, Test

__all__ = ['profile']

X_TEST_SUITE = core.PIGLIT_CONFIG.required_get('xts', 'path')


class XTSProfile(TestProfile):  # pylint: disable=too-few-public-methods
    """ A subclass of TestProfile that provides a setup hook for XTS """

    def setup(self):
        """This hook sets the XTSTest.results_path variable.

        Setting this variable allows images created by XTS to moved into the
        results directory.
        """
        XTSTest.RESULTS_PATH = self.results_dir

        try:
            os.mkdir(os.path.join(self.results_dir, 'images'))
        except OSError as e:
            # If the exception is not 'directory already exists', raise the
            # exception
            if e.errno != 17:
                raise


class XTSTest(Test):  # pylint: disable=too-few-public-methods
    """ X Test Suite class

    Runs a single test or subtest from XTS.

    Arguments:
    name -- the name of the test
    testname -- the name of the test file
    testnum -- the number of the test file

    """
    RESULTS_PATH = None

    def __init__(self, name, testname, testdir, testnum):
        super(XTSTest, self).__init__(
            ['./' + os.path.basename(name), '-i', str(testnum)])
        # Path relative to XTSTest.RESULTS_PATH (which is not
        # initialized at init time) to store any test-specific files.
        # We need to store into the results directory to protect
        # against races when multiple piglit-run.py -t xts commands
        # are running (as in the X Server's make check).
        self.testdir = testdir
        self.testname = '{0}-{1}'.format(testname, testnum)
        self.cwd = os.path.dirname(os.path.realpath(name))
        self.env.update(
            {"XT_RESET_DELAY": '0',
             "XT_FONTPATH_GOOD": '/usr/share/fonts/X11/misc',
             "XT_FONTPATH": os.path.join(X_TEST_SUITE, 'xts5', 'fonts'),
             # XXX: Are the next 3 necissary?
             "XT_LOCAL": 'Yes',
             "XT_TCP": 'No',
             "XT_DISPLAYHOST": ''})

    def run(self):
        # We only get the RESULTS_PATH after the profile has been set
        # up, so we can't do it in init.
        self.test_results_file = os.path.join(XTSTest.RESULTS_PATH,
                                              self.testdir,
                                              self.testname)
        self.env.update({"TET_RESFILE": self.test_results_file})
        super(XTSTest, self).run()

    def _process_log_for_images(self, log):
        """ Parse the image logfile """
        images = []
        search = re.compile('See file (Err[0-9]+.err)')

        for line in log.splitlines():
            match = search.search(line)
            if match is not None:
                # Can we parse any other useful information out to give a
                # better description of each image?
                desc = match.group(1)

                # The error logs are text, with a header with width, height,
                # and depth, then run-length-encoded pixel values (in
                # hexadecimal).  Use xtsttopng to convert the error log to a
                # pair of PNGs so we can put them in the summary.
                command = ['xtsttopng', os.path.join(XTSTest.RESULTS_PATH,
                                                     self.testdir,
                                                     match.group(1))]
                try:
                    out = subprocess.check_output(command, cwd=self.cwd)
                except OSError:
                    images.append({'image_desc': 'image processing failed'})
                    continue

                # Each Err*.err log contains a rendered image, and a reference
                # image that it was compared to.  We relocate the to our tree
                # with more useful names.  (Otherwise, since tests generate
                # error logs with numbers sequentially starting from 0, each
                # subtest with an error would overwrite the previous test's
                # images).
                ref_path = os.path.join(
                    self.RESULTS_PATH, 'images', '{1}-{2}-ref.png'.format(
                        self.testname, match.group(1)))
                render_path = os.path.join(
                    self.RESULTS_PATH, 'images', '{1}-{2}-render.png'.format(
                        self.testname, match.group(1)))

                split = out.splitlines()
                os.rename(os.path.join(self.cwd, split[0]), render_path)
                os.rename(os.path.join(self.cwd, split[1]), ref_path)

                images.append({'image_desc': desc,
                               'image_ref': ref_path,
                               'image_render': render_path})

        return images

    def interpret_result(self):
        super(XTSTest, self).interpret_result()

        try:
            with open(self.test_results_file, 'r') as rfile:
                log = rfile.read()
                self.result.out = log
                os.remove(self.test_results_file)
        except IOError:
            self.result.err = "No results file found"
            log = ""

        if self.result.returncode == 0:
            if re.search('FAIL', self.result.out) is not None:
                self.result.result = 'fail'
            elif re.search('PASS', self.result.out) is not None:
                self.result.result = 'pass'
            else:
                self.result.result = 'fail'
        elif self.result.returncode == 77:
            self.result.result = 'skip'
        elif self.result.returncode == 1:
            if re.search('Could not open all VSW5 fonts', log):
                self.result.result = 'warn'
            else:
                self.result.result = 'fail'

        self.result.images = self._process_log_for_images(log)


class RendercheckTest(Test):
    def __init__(self, args):
        super(RendercheckTest, self).__init__(['rendercheck'] + args)
        self.testname = "rendercheck " + " ".join(args)

    def interpret_result(self):
        super(RendercheckTest, self).interpret_result()

        if self.result.returncode == 0:
            self.result.result = 'pass'
        elif self.result.returncode == 77:
            self.result.result = 'skip'


def _populate_profile_xts(profile):
    fpath = os.path.join(X_TEST_SUITE, 'xts5')
    for dirpath, _, filenames in os.walk(fpath):
        for fname in filenames:
            # only look at the .m test files
            testname, ext = os.path.splitext(fname)
            if ext != '.m':
                continue

            # incrementing number generator
            counts = itertools.count(start=1)

            # Walk the file looking for >>ASSERTION, each of these corresponds
            # to a generated subtest, there can be multiple subtests per .m
            # file
            with open(os.path.join(dirpath, fname), 'r') as rfile:
                for line in rfile:
                    if line.startswith('>>ASSERTION'):
                        num = next(counts)
                        group = grouptools.join(
                            grouptools.from_path(
                                os.path.relpath(dirpath, X_TEST_SUITE)),
                            testname, str(num))

                        profile.test_list[group] = XTSTest(
                            os.path.join(dirpath, testname),
                            testname,
                            os.path.relpath(dirpath, X_TEST_SUITE),
                            num)


def _add_rendercheck_test(profile, path, args):
    test = RendercheckTest(args.split(' '))
    group_path = 'rendercheck/' + path
    group = grouptools.join(*(group_path.split('/')))
    profile.test_list[group] = test


def _populate_profile_rendercheck(profile):
    _add_rendercheck_test(profile, 'blend/All/a8r8g8b8', '-t blend -f a8r8g8b8')
    _add_rendercheck_test(profile, 'blend/All/x8r8g8b8', '-t blend -f a8r8g8b8,x8r8g8b8')
    _add_rendercheck_test(profile, 'blend/All/a2r10g10b10', '-t blend -f a8r8g8b8,a2r10g10b10')
    _add_rendercheck_test(profile, 'blend/Clear', '-t blend -o clear')
    _add_rendercheck_test(profile, 'blend/Src', '-t blend -o src')
    _add_rendercheck_test(profile, 'blend/Over', '-t blend -o over')
    _add_rendercheck_test(profile, 'composite/All/a8r8g8b8', '-t composite -f a8r8g8b8')
    _add_rendercheck_test(profile, 'composite/All/x8r8g8b8', '-t composite -f a8r8g8b8,x8r8g8b8')
    _add_rendercheck_test(profile, 'composite/All/a2r10g10b10', '-t composite -f a8r8g8b8,a2r10g10b10')
    _add_rendercheck_test(profile, 'ca composite/All/a8r8g8b8', '-t cacomposite -f a8r8g8b8')
    _add_rendercheck_test(profile, 'ca composite/All/a8', '-t cacomposite -f a8r8g8b8,a8')
    _add_rendercheck_test(profile, 'ca composite/All/x8r8g8b8', '-t cacomposite -f a8r8g8b8,x8r8g8b8')
    _add_rendercheck_test(profile, 'ca composite/All/a2r10g10b10', '-t cacomposite -f a8r8g8b8,a2r10g10b10')
    _add_rendercheck_test(profile, 'fill', '-t fill')
    _add_rendercheck_test(profile, 'bug7366', '-t bug7366')
    _add_rendercheck_test(profile, 'destination coordinates', '-t dcoords')
    _add_rendercheck_test(profile, 'source coordinates', '-t scoords')
    _add_rendercheck_test(profile, 'mask coordinates', '-t mcoords')
    _add_rendercheck_test(profile, 'translated source coordinates', '-t tscoords')
    _add_rendercheck_test(profile, 'translated mask coordinates', '-t tmcoords')
    _add_rendercheck_test(profile, 'triangles', '-t triangles')
    _add_rendercheck_test(profile, 'LibreOffice xRGB', '-t libreoffice_xrgb')
    _add_rendercheck_test(profile, 'GTK ARGB vs xBGR', '-t gtk_argb_xbgr')


def _populate_profile():
    """ Populate the profile attribute """
    # Add all tests to the profile
    profile = XTSProfile()  # pylint: disable=redefined-outer-name
    _populate_profile_xts(profile)
    _populate_profile_rendercheck(profile)
    return profile


if not os.path.exists(X_TEST_SUITE):
    raise exceptions.PiglitFatalError('XTest suite not found.')

profile = _populate_profile()  # pylint: disable=invalid-name



# encoding=utf-8
# Copyright © 2018 Intel Coproration

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import random

from framework import grouptools
from .shader import profile as _profile

__all__ = ['profile']

profile = _profile.copy()


# There are too many of theset, it's simply exhaustive to run them all.
class FilterVsIn(object):
    """Filter out 80% of the Vertex Attrib 64 vs_in tests."""

    def __init__(self):
        self.random = random.Random()
        self.random.seed(42)

    def __call__(self, name, _):
        if 'vs_in' in grouptools.split(name):
            # 20%
            return self.random.random() <= .2
        return True

    def reset(self):
        self.random.seed(42)


profile.filters.append(FilterVsIn())

# These take too long
profile.filters.append(lambda n, _: '-explosion' not in n)



# encoding=utf-8
# Copyright © 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for generated_tests/modules/glsl.py."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import itertools
import operator

import pytest
import six

# pylint can't figure out the sys.path manipulation.
from modules import glsl  # pylint: disable=import-error

# pylint: disable=no-self-use,invalid-name

_GLSL = [
    '110',
    '120',
    '130',
    '140',
    '150',
    '330',
    '400',
    '410',
    '420',
    '430',
    '440',
    '450',
]

_GLSLES = [
    '100',
    '300 es',
    '310 es',
    '320 es',
]

_OPERATORS = [
    operator.lt,
    operator.le,
    operator.eq,
    operator.ne,
    operator.ge,
    operator.gt,
]


def test_factory_glsl():
    """generated_tests.modules.glsl.Version: provides a GLSLVersion."""
    test = glsl.Version('110')
    assert isinstance(test, glsl.GLSLVersion)


def test_factory_glsles():
    """generated_tests.modules.glsl.Version: provides a GLSLESVersion."""
    test = glsl.Version('100')
    assert isinstance(test, glsl.GLSLESVersion)


def test_factory_cache():
    """generated_tests.modules.glsl.Version: caches objects."""
    test1 = glsl.Version('100')
    test2 = glsl.Version('100')
    assert test1 is test2


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSL, _GLSL))
def test_compare_glsl_to_glsl(op, first, second):
    """Test GLSLVersion <cmp> GLSLVersion."""
    actual = op(glsl.Version(first), glsl.Version(second))
    assert actual == op(int(first), int(second))


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSLES, _GLSLES))
def test_compare_glsles_to_glsles(op, first, second):
    """Test GLSLESVersion <cmp> GLSLESVersion."""
    actual = op(glsl.Version(first), glsl.Version(second))
    assert actual == op(int(first[:3]), int(second[:3]))


@pytest.mark.raises(exception=TypeError)
def test_glsl_glsles_exception():
    """generated_tests.modules.glsl: GLSLVersion <cmp> GLSLESVersion."""
    return glsl.Version('110') < glsl.Version('100')


@pytest.mark.raises(exception=TypeError)
def test_glsles_glsl_exception():
    """generated_tests.modules.glsl: GLSLESVersion <cmp> GLSLVersion."""
    return glsl.Version('100') < glsl.Version('110')


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSL, _GLSL))
def test_compare_glsl_to_int(op, first, second):
    """Test GLSLVersion <cmp> GLSLVersion."""
    actual = op(glsl.Version(first), int(second))
    assert actual == op(int(first), int(second))


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSLES, _GLSLES))
def test_compare_glsles_to_int(op, first, second):
    """Test GLSLESVersion <cmp> GLSLESVersion."""
    actual = op(glsl.Version(first), int(second[:3]))
    assert actual == op(int(first[:3]), int(second[:3]))


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSL, _GLSL))
def test_compare_glsl_to_float(op, first, second):
    """Test GLSLVersion <cmp> GLSLVersion."""
    actual = op(glsl.Version(first), float(second) / 100)
    expected = op(float(first), float(second))
    assert actual == expected


@pytest.mark.parametrize("op,first,second",
                         itertools.product(_OPERATORS, _GLSLES, _GLSLES))
def test_compare_glsles_to_float(op, first, second):
    """Test GLSLESVersion <cmp> GLSLESVersion."""
    actual = op(glsl.Version(first), float(second[:3]) / 100)
    assert actual == op(float(first[:3]), float(second[:3]))




def test_GLSLVersion_str():
    """generated_tests.modules.glsl.GLSLVersion: str()"""
    assert six.text_type(glsl.Version('110')) == '110'


def test_GLSLESVersion_str():
    """generated_tests.modules.glsl.GLSLESVersion: str()"""
    assert six.text_type(glsl.Version('100')) == '100'


def test_GLSLVersion_int():
    """generated_tests.modules.glsl.GLSLVersion: int()"""
    assert int(glsl.Version('110')) == 110


def test_GLSLESVersion_int():
    """generated_tests.modules.glsl.GLSLESVersion: int()"""
    assert int(glsl.Version('100')) == 100


def test_GLSLVersion_float():
    """generated_tests.modules.glsl.GLSLVersion: float()"""
    assert float(glsl.Version('110')) == 1.10


def test_GLSLESVersion_float():
    """generated_tests.modules.glsl.GLSLESVersion: float()"""
    assert float(glsl.Version('100')) == 1.00


def test_GLSLVersion_print_float():
    """generated_tests.modules.glsl.GLSLVersion: print_float()"""
    assert glsl.Version('110').print_float() == '1.10'


def test_GLSLESVersion_print_float():
    """generated_tests.modules.glsl.GLSLESVersion: print_float()"""
    assert glsl.Version('100').print_float() == '1.00'


def test_GLSLESVersion_print_float_es():
    """generated_tests.modules.glsl.GLSLESVersion: print_float() (es version)"""
    assert glsl.Version('300 es').print_float() == '3.00 es'


class TestMinVersion_for_stage(object):
    """Tests for generated_tests.modules.glsl.MinVersion.for_stage.

    Each test covers the requested < required and requested == required. If
    it's possible it also covers requested > required.

    """
    def _test(self, stage, version, expected):
        assert glsl.MinVersion.for_stage(stage, version) == expected

    def test_opengl_frag(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: FS (OpenGL)"""
        self._test('frag', glsl.Version('150'), glsl.Version('150'))
        self._test('frag', glsl.Version('110'), glsl.Version('110'))

    def test_opengl_vert(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: VS (OpenGL)"""
        self._test('vert', glsl.Version('150'), glsl.Version('150'))
        self._test('vert', glsl.Version('110'), glsl.Version('110'))

    def test_opengl_geom(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: GS (OpenGL)"""
        self._test('geom', glsl.Version('330'), glsl.Version('330'))
        self._test('geom', glsl.Version('110'), glsl.Version('150'))
        self._test('geom', glsl.Version('150'), glsl.Version('150'))

    def test_opengl_tesc(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: TCS (OpenGL)"""
        self._test('tesc', glsl.Version('410'), glsl.Version('410'))
        self._test('tesc', glsl.Version('400'), glsl.Version('400'))
        self._test('tesc', glsl.Version('140'), glsl.Version('400'))

    def test_opengl_tese(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: TES (OpenGL)"""
        self._test('tese', glsl.Version('410'), glsl.Version('410'))
        self._test('tese', glsl.Version('400'), glsl.Version('400'))
        self._test('tese', glsl.Version('140'), glsl.Version('400'))

    def test_opengl_comp(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: CS (OpenGL)"""
        self._test('comp', glsl.Version('440'), glsl.Version('440'))
        self._test('comp', glsl.Version('430'), glsl.Version('430'))
        self._test('comp', glsl.Version('140'), glsl.Version('430'))

    def test_opengles_frag(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: FS (OpenGL ES)"""
        self._test('frag', glsl.Version('300 es'), glsl.Version('300 es'))
        self._test('frag', glsl.Version('100'), glsl.Version('100'))

    def test_opengles_vert(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: VS (OpenGL ES)"""
        self._test('vert', glsl.Version('320 es'), glsl.Version('320 es'))
        self._test('vert', glsl.Version('100'), glsl.Version('100'))

    def test_opengles_geom(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: GS (OpenGL ES)"""
        self._test('geom', glsl.Version('100'), glsl.Version('320 es'))
        self._test('geom', glsl.Version('320 es'), glsl.Version('320 es'))

    def test_opengles_tesc(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: TCS (OpenGL ES)"""
        self._test('tesc', glsl.Version('320 es'), glsl.Version('320 es'))
        self._test('tesc', glsl.Version('100'), glsl.Version('320 es'))

    def test_opengles_tese(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: TES (OpenGL ES)"""
        self._test('tese', glsl.Version('320 es'), glsl.Version('320 es'))
        self._test('tese', glsl.Version('100'), glsl.Version('320 es'))

    def test_opengles_comp(self):
        """generated_tests.modules.glsl.MinVersion.for_stage: TES (OpenGL ES)"""
        self._test('comp', glsl.Version('320 es'), glsl.Version('320 es'))
        self._test('comp', glsl.Version('100'), glsl.Version('310 es'))
        self._test('comp', glsl.Version('310 es'), glsl.Version('310 es'))


class TestMinVersion_for_stage_with_ext(object):
    """Tests for generated_tests.modules.glsl.MinVersion.for_stage_with_ext."""
    def _test(self, stage, version, expected):
        ver, ext = glsl.MinVersion.for_stage_with_ext(stage, version)
        assert (ver, ext) == expected

    def test_opengl_frag(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        FS (OpenGL)"""
        self._test('frag', glsl.Version('150'), (glsl.Version('150'), None))
        self._test('frag', glsl.Version('110'), (glsl.Version('110'), None))

    def test_opengl_vert(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        VS (OpenGL)"""
        self._test('vert', glsl.Version('150'), (glsl.Version('150'), None))
        self._test('vert', glsl.Version('110'), (glsl.Version('110'), None))

    def test_opengl_geom(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        GS (OpenGL)"""
        self._test('geom', glsl.Version('330'), (glsl.Version('330'), None))
        self._test('geom', glsl.Version('110'), (glsl.Version('150'), None))
        self._test('geom', glsl.Version('150'), (glsl.Version('150'), None))

    def test_opengl_tesc(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        TCS (OpenGL)"""
        self._test('tesc', glsl.Version('410'), (glsl.Version('410'), None))
        self._test('tesc', glsl.Version('140'),
                   (glsl.Version('140'), 'GL_ARB_tessellation_shader'))
        self._test('tesc', glsl.Version('110'),
                   (glsl.Version('140'), 'GL_ARB_tessellation_shader'))

    def test_opengl_tese(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        TES (OpenGL)"""
        self._test('tese', glsl.Version('410'), (glsl.Version('410'), None))
        self._test('tese', glsl.Version('140'),
                   (glsl.Version('140'), 'GL_ARB_tessellation_shader'))
        self._test('tese', glsl.Version('110'),
                   (glsl.Version('140'), 'GL_ARB_tessellation_shader'))

    def test_opengl_comp(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        CS (OpenGL)"""
        self._test('comp', glsl.Version('430'), (glsl.Version('430'), None))
        self._test('comp', glsl.Version('140'),
                   (glsl.Version('140'), 'GL_ARB_compute_shader'))
        self._test('comp', glsl.Version('110'),
                   (glsl.Version('140'), 'GL_ARB_compute_shader'))

    def test_opengles_frag(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        FS (OpenGL ES)"""
        self._test('frag', glsl.Version('300 es'),
                   (glsl.Version('300 es'), None))
        self._test('frag', glsl.Version('100'), (glsl.Version('100'), None))

    def test_opengles_vert(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        VS (OpenGL ES)"""
        self._test('vert', glsl.Version('300 es'),
                   (glsl.Version('300 es'), None))
        self._test('vert', glsl.Version('100'), (glsl.Version('100'), None))

    def test_opengles_geom(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        GS (OpenGL ES)"""
        self._test('geom', glsl.Version('100'),
                   (glsl.Version('310 es'), 'GL_OES_geometry_shader'))
        self._test('geom', glsl.Version('310 es'),
                   (glsl.Version('310 es'), 'GL_OES_geometry_shader'))
        self._test('geom', glsl.Version('320 es'),
                   (glsl.Version('320 es'), None))

    def test_opengles_tesc(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        TCS (OpenGL ES)"""
        self._test('tesc', glsl.Version('320 es'),
                   (glsl.Version('320 es'), None))
        self._test('tesc', glsl.Version('310 es'),
                   (glsl.Version('310 es'), 'GL_OES_tessellation_shader'))
        self._test('tesc', glsl.Version('100'),
                   (glsl.Version('310 es'), 'GL_OES_tessellation_shader'))

    def test_opengles_tese(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        TES (OpenGL ES)"""
        self._test('tese', glsl.Version('320 es'),
                   (glsl.Version('320 es'), None))
        self._test('tese', glsl.Version('310 es'),
                   (glsl.Version('310 es'), 'GL_OES_tessellation_shader'))
        self._test('tese', glsl.Version('100'),
                   (glsl.Version('310 es'), 'GL_OES_tessellation_shader'))

    def test_opengles_comp(self):
        """generated_tests.modules.glsl.MinVersion.for_stage_with_ext:
        TES (OpenGL ES)"""
        self._test('comp', glsl.Version('320 es'),
                   (glsl.Version('320 es'), None))
        self._test('comp', glsl.Version('100'), (glsl.Version('310 es'), None))
        self._test('comp', glsl.Version('310 es'),
                   (glsl.Version('310 es'), None))



# coding=utf-8
#
# Copyright © 2011, 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Generate Interpopation tests.

Correct interpolation of vertex shader outputs depends on (a)
whether an interpolation qualifier is present in the shader source,
and if so which qualifier is used, (b) if no interpolation qualifier
is present, whether the output is a user-defined variable or a
built-in color, and (c) if the output is a built-in color, the
setting of the ShadeModel() setting.  In addition, we would like to
test correct interpolation under various clipping scenarios.

To verify that all the combinations of these possibilities work
correctly, this script generates a shader_runner test to check
proper interpolation for every combination of the following
variables:

- which interpolation qualifier is used ("flat", "noperspective",
  "smooth", or no qualifier)

- which variable is used (gl_FrontColor, gl_BackColor,
  gl_FrontSecondaryColor, gl_BackSecondaryColor, or a non-built-in
  variable)

- the setting of ShadeModel() (either GL_SMOOTH or GL_FLAT)

- whether the triangle in question is clipped using gl_ClipVertex,
  clipped using gl_ClipDistance, clipped against the fixed viewing
  volume, or unclipped.

The tests operate by drawing a triangle with a different value of
the variable at each vertex, and then probing within the interior of
the triangle to verify that interpolation was performed correctly.
The triangle is drawn in a frustum projection, with a different z
value for each vertex, so that there will be a detectable difference
in behavior between noperspective and smooth interpolation.

When testing clipping, we clip off the frontmost corner of the
triangle; this ensures that the proportion of the triangle's screen
real estate that is clipped is significantly larger than the
proportion of the triangle's 3D coordinate space that is clipped.
So if the GL implementation doesn't perform perspective-correct
interpolation generating clipped vertices, we will notice.

This program outputs, to stdout, the name of each file it generates.

"""

from __future__ import print_function, division, absolute_import
import os

from six.moves import range

from templates import template_file
from modules import utils

TEMPLATE = template_file(os.path.basename(os.path.splitext(__file__)[0]),
                         'template.shader_test.mako')


class Test(object):
    def __init__(self, interpolation_qualifier, variable, shade_model,
                 clipping):
        """Get ready to generate a test using the given settings.

        interpolation_qualifier is a string representing the desired
        interpolation qualifier that should appear in GLSL code
        ('flat', 'noperspective', or 'smooth'), or None if no
        qualifier should appear.

        variable is the name of the variable on which to test
        interpolation.  If the name begins with 'gl_', it should be
        one of the four vertex shader variables that are allowed to be
        redeclared with an interpolation qualifier (see GLSL 1.30
        section 4.3.7 "Interpolation").  Namely: gl_FrontColor,
        gl_BackColor, gl_FrontSecondaryColor, or
        gl_BackSecondaryColor.

        shade_model is which shade model the GL state should be put in
        using the glShadeModel() command--either 'smooth' or 'flat'.

        clipping is the variety of clipping which should be tested:
        either 'fixed' to test a triangle that extends beyond the
        fixed view volume (we test clipping against the "near" plane),
        'vertex' to test a triangle which has one corner clipped using
        gl_ClipVertex, or 'distance' to test a triangle which has one
        corner clipped using gl_ClipDistance.
        """
        self.interpolation_qualifier = interpolation_qualifier
        self.vs_variable = variable
        self.shade_model = shade_model
        self.clipping = clipping

        # When colors are mapped into the fragment shader, the string
        # 'Front' or 'Back' is dropped from the variable name, since
        # e.g. gl_Color is mapped to gl_FrontColor for front-facing
        # triangles, and gl_BackColor for back-facing triangles.
        self.fs_variable = variable.replace('Front', '').replace('Back', '')

        # True if we are testing a BackColor, so we'll need to draw a
        # back-facing triangle.
        self.backfacing = variable.find('Back') != -1

        # True if we are testing a built-in color variable, False if
        # we are testing a generic vertex shader output.
        self.builtin_variable = variable[:3] == 'gl_'

        # Determine whether the test requires GLSL 1.30.  If it does,
        # use "in" and "out" to qualify shader inputs and outputs.
        # Otherwise use the old keywords "attribute" and "varying".
        # shader_runner will insert a #version directive based on
        # glsl_version.
        if self.interpolation_qualifier or self.clipping == 'distance':
            self.glsl_version = '1.30'
            self.vs_input = 'in'
            self.vs_output = 'out'
            self.fs_input = 'in'
        else:
            self.glsl_version = '1.10'
            self.vs_input = 'attribute'
            self.vs_output = 'varying'
            self.fs_input = 'varying'

        # Determine the location of the near and far planes for the
        # frustum projection.  The triangle fits between z coordinates
        # -1 and -3; we use 1.75 as the near plane when we want to
        # force clipping.
        if self.clipping == 'fixed':
            self.frustum_near = 1.75
        else:
            self.frustum_near = 1.0
        self.frustum_far = 3.0

        # Determine whether we expect the GL implementation to use
        # flatshading, non-perspective interpolation, or perspective
        # interpolation.
        if self.interpolation_qualifier:
            # According to GLSL 1.30 section 4.3.7 ("Interpolation"),
            # "When an interpolation qualifier is used, it overrides
            # settings established through the OpenGL API."
            self.expected_behavior = self.interpolation_qualifier
        elif self.builtin_variable:
            # According to GL 3.0 section 2.19.7 ("Flatshading"), "If
            # a vertex shader is active, the flat shading control
            # applies to the built-in varying variables gl FrontColor,
            # gl BackColor, gl FrontSecondaryColor and gl
            # BackSecondaryColor.  Non-color varying variables can be
            # specified as being flat-shaded via the flat qualifier,
            # as described in section 4.3.6 of the OpenGL Shading
            # Language Specification."
            self.expected_behavior = self.shade_model
        else:
            # The specs do not explicitly state how non-built-in
            # variables are to be interpolated in the case where no
            # interpolation qualifier is used.  However, it seems to
            # be heavily implied by the text of GL 3.0 section 2.19.6
            # ("Flatshading"--see above) that smooth
            # (perspective-correct) interpolation is intended,
            # regardless of the setting of glShadeModel().
            self.expected_behavior = 'smooth'

    def filename(self):
        return os.path.join(
            'spec', 'glsl-{0}'.format(self.glsl_version),
            'execution', 'interpolation',
            'interpolation-{0}-{1}-{2}-{3}.shader_test'.format(
                self.interpolation_qualifier or 'none', self.vs_variable,
                self.shade_model, self.clipping or 'none'))

    def vertex_data(self):
        table = ['vertex/float/3  input_data/float/4',
                 '-1.0 -1.0 -1.0  1.0 0.0 0.0 1.0',
                 ' 0.0  2.0 -2.0  0.0 1.0 0.0 1.0',
                 ' 3.0 -3.0 -3.0  0.0 0.0 1.0 1.0']
        if not self.backfacing:
            # The vertices above are ordered such that the front of
            # the triangle faces away from the viewer.  If we are
            # trying to render the front face, then swap the first two
            # vertices.  This shows us the front face of the triangle
            # without changing the provoking vertex (which is the
            # third vertex).
            table = [table[0], table[2], table[1], table[3]]
        return table

    def probe_data(self):
        # Loop over possible barycentric coordinates with a spacing of
        # 1/num_subdivisions.  Skip points on the triangle edges and
        # corners so that rounding does not cause us to accidentally
        # probe a pixel that's outside the triangle.
        num_subdivisions = 6
        for i in range(1, num_subdivisions - 1):
            for j in range(1, num_subdivisions - i):
                # Compute 3D barycentric coordinates--these will be
                # used to compute the expected interpolated values
                # when using smooth (perspective-correct)
                # interpolation.  The vertex associated with b3d_0=1.0
                # is colored red, the vertex associated with b3d_1=1.0
                # is colored green, and the vertex associated with
                # b3d_2=1.0 is colored blue.
                b3d_0 = float(num_subdivisions - i - j)/num_subdivisions
                b3d_1 = float(i)/num_subdivisions
                b3d_2 = float(j)/num_subdivisions
                # Compute 3D coordinates based on those barycentric
                # coordinates.  These will be used, among other
                # things, to determine whether this part of the
                # triangle is clipped.
                x3d = -b3d_0 + 3.0*b3d_2
                y3d = -b3d_0 + 2.0*b3d_1 - 3.0*b3d_2
                z3d = -b3d_0 - 2.0*b3d_1 - 3.0*b3d_2
                # Use perspective division to compute 2D screen
                # coordinates.  These will be used with "relative
                # probe rgba", which treats the lower left corner of
                # the screen as (0, 0) and the upper right is (1, 1).
                x2d = (-x3d/z3d + 1.0) / 2.0
                y2d = (-y3d/z3d + 1.0) / 2.0
                # Finally, compute a second set of barycentric
                # coordinates based on the 2D screen
                # coordinates--these will be used to compute the
                # expected interpolated values when using
                # noperspective (screen-coordinate) interpolation.
                b2d_0 = 1.0 - x2d - 0.5*y2d
                b2d_1 = y2d
                b2d_2 = x2d - 0.5*y2d

                if self.clipping and -z3d < 1.75:
                    # Points whose -z coordinate is less than 1.75
                    # should be clipped.
                    yield x2d, y2d, 0.0, 0.0, 0.0, 0.0
                elif self.expected_behavior == 'flat':
                    # When flatshading, all points on the triangle
                    # should inherit the color of the third vertex,
                    # which is blue.
                    yield x2d, y2d, 0.0, 0.0, 1.0, 1.0
                elif self.expected_behavior == 'noperspective':
                    # Since the 3 triangle vertices are red, green,
                    # and blue, the interpolated color channels should
                    # be exactly equal to the barycentric coordinates.
                    # For "noperspective" shading, we use the
                    # barycentric coordinates that we computed based
                    # on 2D screen position.
                    yield x2d, y2d, b2d_0, b2d_1, b2d_2, 1.0
                else:
                    # For "smooth" (perspective correct) shading, we
                    # use the barycentric coordinates that we used to
                    # compute the 3D position.
                    assert self.expected_behavior == 'smooth'
                    yield x2d, y2d, b3d_0, b3d_1, b3d_2, 1.0

    def generate(self):
        filename = self.filename()
        dirname = os.path.dirname(filename)
        utils.safe_makedirs(dirname)
        with open(filename, 'w') as f:
            f.write(TEMPLATE.render_unicode(args=self))


def all_tests():
    for interpolation_qualifier in ['flat', 'smooth', 'noperspective', '']:
        for variable in ['gl_FrontColor', 'gl_BackColor',
                         'gl_FrontSecondaryColor', 'gl_BackSecondaryColor',
                         'other']:
            for shade_model in ['smooth', 'flat']:
                for clipping in ['vertex', 'distance', 'fixed', '']:
                    yield Test(interpolation_qualifier, variable, shade_model,
                               clipping)


def main():
    for test in all_tests():
        test.generate()
        print(test.filename())


if __name__ == '__main__':
    main()
